"""Resource for target groups"""
from typing import Any, Dict, List, Type

from botocore.client import BaseClient
from botocore.exceptions import ClientError

from altimeter.aws.resource.resource_spec import ListFromAWSResult
from altimeter.aws.resource.ec2.vpc import VPCResourceSpec
from altimeter.aws.resource.elbv2 import ELBV2ResourceSpec
from altimeter.aws.resource.elbv2.load_balancer import LoadBalancerResourceSpec
from altimeter.core.exceptions import AltimeterException
from altimeter.core.graph.field.dict_field import AnonymousDictField, EmbeddedDictField
from altimeter.core.graph.field.list_field import ListField
from altimeter.core.graph.field.resource_link_field import (
    EmbeddedResourceLinkField,
    TransientResourceLinkField,
)
from altimeter.core.graph.field.scalar_field import ScalarField
from altimeter.core.graph.schema import Schema


class TargetGroupAccessDeniedException(AltimeterException):
    pass


class TargetGroupResourceSpec(ELBV2ResourceSpec):
    """Resource for target group"""

    type_name = "targetgroup"
    schema = Schema(
        ScalarField("TargetGroupName"),
        ScalarField("Protocol", optional=True),
        ScalarField("Port", optional=True),
        TransientResourceLinkField("VpcId", VPCResourceSpec, optional=True),
        ScalarField("HealthCheckProtocol", optional=True),
        ScalarField("HealthCheckPort", optional=True),
        ScalarField("HealthCheckEnabled"),
        ListField(
            "LoadBalancerArns",
            EmbeddedResourceLinkField(LoadBalancerResourceSpec, value_is_id=True),
        ),
        ScalarField("TargetType"),
        ListField(
            "TargetHealthDescriptions",
            EmbeddedDictField(
                AnonymousDictField(
                    "Target",
                    ScalarField("Id", alti_key="target_id"),
                    ScalarField("Port", alti_key="target_port", optional=True),
                    ScalarField("AvailabilityZone", alti_key="target_az", optional=True),
                ),
                ScalarField("HealthCheckPort", optional=True),
                AnonymousDictField(
                    "TargetHealth",
                    ScalarField("State"),
                    ScalarField("Reason", optional=True),
                    ScalarField("Description", optional=True),
                    optional=True,
                ),
            ),
        ),
    )

    @classmethod
    def list_from_aws(
        cls: Type["TargetGroupResourceSpec"], client: BaseClient, account_id: str, region: str
    ) -> ListFromAWSResult:
        """Return a dict of dicts of the format:

            {'target_group_1_arn': {target_group_1_dict},
             'target_group_2_arn': {target_group_2_dict},
             ...}

        Where the dicts represent results from describe_target_groups."""
        paginator = client.get_paginator("describe_target_groups")
        resources = {}
        for resp in paginator.paginate():
            for resource in resp.get("TargetGroups", []):
                resource_arn = resource["TargetGroupArn"]
                resource["TargetHealthDescriptions"] = get_target_group_health(client, resource_arn)
                resources[resource_arn] = resource
        return ListFromAWSResult(resources=resources)


def get_target_group_health(client: BaseClient, target_group_arn: str) -> List[Dict[str, Any]]:
    """Describes target health for a given target group"""
    try:
        return client.describe_target_health(TargetGroupArn=target_group_arn).get(
            "TargetHealthDescriptions", []
        )
    except ClientError as c_e:
        response_error = getattr(c_e, "response", {}).get("Error", {})
        error_code = response_error.get("Code", "")
        if error_code == "AccessDenied":
            raise TargetGroupAccessDeniedException(
                f"Error getting encryption configuration for {target_group_arn}: {response_error}",
                c_e,
            )
        raise c_e



"""AWS Resource classes."""
from typing import Tuple, Type

from altimeter.aws.resource.resource_spec import AWSResourceSpec
from altimeter.aws.resource.account import AccountResourceSpec
from altimeter.aws.resource.awslambda.function import LambdaFunctionResourceSpec
from altimeter.aws.resource.dynamodb.dynamodb_table import DynamoDbTableResourceSpec
from altimeter.aws.resource.ec2.flow_log import FlowLogResourceSpec
from altimeter.aws.resource.ec2.image import EC2ImageResourceSpec
from altimeter.aws.resource.ec2.instance import EC2InstanceResourceSpec
from altimeter.aws.resource.ec2.network_interface import EC2NetworkInterfaceResourceSpec
from altimeter.aws.resource.ec2.region import RegionResourceSpec
from altimeter.aws.resource.ec2.route_table import EC2RouteTableResourceSpec
from altimeter.aws.resource.ec2.transit_gateway_vpc_attachment import (
    TransitGatewayVpcAttachmentResourceSpec,
)
from altimeter.aws.resource.ec2.security_group import SecurityGroupResourceSpec
from altimeter.aws.resource.ec2.snapshot import EBSSnapshotResourceSpec
from altimeter.aws.resource.ec2.subnet import SubnetResourceSpec
from altimeter.aws.resource.ec2.transit_gateway import TransitGatewayResourceSpec
from altimeter.aws.resource.ec2.volume import EBSVolumeResourceSpec
from altimeter.aws.resource.ec2.vpc import VPCResourceSpec
from altimeter.aws.resource.ec2.vpc_endpoint import VpcEndpointResourceSpec
from altimeter.aws.resource.elbv1.load_balancer import ClassicLoadBalancerResourceSpec
from altimeter.aws.resource.elbv2.load_balancer import LoadBalancerResourceSpec
from altimeter.aws.resource.elbv2.target_group import TargetGroupResourceSpec
from altimeter.aws.resource.eks.cluster import EKSClusterResourceSpec
from altimeter.aws.resource.events.cloudwatchevents_rule import EventsRuleResourceSpec
from altimeter.aws.resource.events.event_bus import EventBusResourceSpec
from altimeter.aws.resource.guardduty.detector import DetectorResourceSpec
from altimeter.aws.resource.iam.account_password_policy import IAMAccountPasswordPolicyResourceSpec
from altimeter.aws.resource.iam.iam_saml_provider import IAMSAMLProviderResourceSpec
from altimeter.aws.resource.iam.instance_profile import InstanceProfileResourceSpec
from altimeter.aws.resource.iam.policy import IAMPolicyResourceSpec, IAMAWSManagedPolicyResourceSpec
from altimeter.aws.resource.iam.role import IAMRoleResourceSpec
from altimeter.aws.resource.iam.user import IAMUserResourceSpec
from altimeter.aws.resource.kms.key import KMSKeyResourceSpec
from altimeter.aws.resource.organizations.org import OrgResourceSpec
from altimeter.aws.resource.organizations.ou import OUResourceSpec
from altimeter.aws.resource.organizations.account import OrgsAccountResourceSpec
from altimeter.aws.resource.rds.instance import RDSInstanceResourceSpec
from altimeter.aws.resource.rds.snapshot import RDSSnapshotResourceSpec
from altimeter.aws.resource.s3.bucket import S3BucketResourceSpec
from altimeter.aws.resource.support.severity_level import SeverityLevelResourceSpec

# To enable a resource to be scanned, add it here.
RESOURCE_SPEC_CLASSES: Tuple[Type[AWSResourceSpec], ...] = (
    ClassicLoadBalancerResourceSpec,
    DetectorResourceSpec,
    DynamoDbTableResourceSpec,
    EBSSnapshotResourceSpec,
    EBSVolumeResourceSpec,
    EC2ImageResourceSpec,
    EC2InstanceResourceSpec,
    EC2NetworkInterfaceResourceSpec,
    EC2RouteTableResourceSpec,
    EKSClusterResourceSpec,
    EventBusResourceSpec,
    EventsRuleResourceSpec,
    FlowLogResourceSpec,
    IAMAccountPasswordPolicyResourceSpec,
    IAMAWSManagedPolicyResourceSpec,
    IAMPolicyResourceSpec,
    IAMRoleResourceSpec,
    IAMSAMLProviderResourceSpec,
    IAMUserResourceSpec,
    InstanceProfileResourceSpec,
    KMSKeyResourceSpec,
    LambdaFunctionResourceSpec,
    LoadBalancerResourceSpec,
    RDSInstanceResourceSpec,
    RDSSnapshotResourceSpec,
    S3BucketResourceSpec,
    SecurityGroupResourceSpec,
    SeverityLevelResourceSpec,
    SubnetResourceSpec,
    TargetGroupResourceSpec,
    TransitGatewayResourceSpec,
    TransitGatewayVpcAttachmentResourceSpec,
    VPCResourceSpec,
    VpcEndpointResourceSpec,
)

INFRA_RESOURCE_SPEC_CLASSES: Tuple[Type[AWSResourceSpec], ...] = (
    AccountResourceSpec,
    RegionResourceSpec,
)

ORG_RESOURCE_SPEC_CLASSES: Tuple[Type[AWSResourceSpec], ...] = (
    OrgResourceSpec,
    OrgsAccountResourceSpec,
    OUResourceSpec,
)



"""Base class for Support resources."""
from altimeter.aws.resource.resource_spec import ScanGranularity, AWSResourceSpec


class SupportResourceSpec(AWSResourceSpec):
    """Base class for Support resources."""

    service_name = "support"
    scan_granularity = ScanGranularity.ACCOUNT
    region_whitelist = ("us-east-1",)



"""A NodeCache is a simple cache for graph nodes."""
from rdflib import BNode


class NodeCache(dict):
    """A NodeCache is a simple cache for graph nodes. Unlike a standard
    dict it does not allow overwriting entries."""

    def __setitem__(self, key: str, value: BNode) -> None:
        """Set an item in this NodeCache
            key: cache key
            value: BNode value

        Raises:
            KeyError if this key is already present.
        """
        if key in self:
            raise KeyError(f"Key already present for {key}")
        super().__setitem__(key, value)



"""List Fields represent fields which consist of list data."""
from copy import deepcopy
from typing import Dict, Any, List

from altimeter.core.graph.exceptions import (
    ListFieldSourceKeyNotFoundException,
    ListFieldValueNotAListException,
)
from altimeter.core.graph.field.base import Field, SubField
from altimeter.core.graph.field.scalar_field import SCALAR_TYPES
from altimeter.core.graph.field.util import camel_case_to_snake_case
from altimeter.core.graph.link.base import Link


class ListField(Field):
    """A ListField is a field where the input is a JSON object containing a key (source_key)
    where the corresponding value is a list of homogeneous items.

    Examples:
        A list of strings:
            >>> from altimeter.core.graph.field.scalar_field import EmbeddedScalarField
            >>> input = {"Animals": ["cow", "pig", "human"]}
            >>> field = ListField("Animals", EmbeddedScalarField())
            >>> links = field.parse(data=input, context={})
            >>> for link in links: print(link.to_dict())
            {'pred': 'animals', 'obj': 'cow', 'type': 'simple'}
            {'pred': 'animals', 'obj': 'pig', 'type': 'simple'}
            {'pred': 'animals', 'obj': 'human', 'type': 'simple'}

        A list of dicts:
            >>> from altimeter.core.graph.field.dict_field import EmbeddedDictField
            >>> from altimeter.core.graph.field.scalar_field import ScalarField
            >>> input = {"People": [{"Name": "Bob", "Age": 49}, {"Name": "Sue", "Age": 42}]}
            >>> field = ListField("People", EmbeddedDictField(ScalarField("Name"),\
                                  ScalarField("Age")))
            >>> links = field.parse(data=input, context={})
            >>> for link in links:
            ...   print(link.pred)
            ...   for obj in link.obj:
            ...      print(obj.to_dict())
            ...
            people
            {'pred': 'name', 'obj': 'Bob', 'type': 'simple'}
            {'pred': 'age', 'obj': 49, 'type': 'simple'}
            people
            {'pred': 'name', 'obj': 'Sue', 'type': 'simple'}
            {'pred': 'age', 'obj': 42, 'type': 'simple'}

    Args:
        source_key: Name of the key in the input JSON
        sub_field: SubField object representing the type that is contained in this list.
        alti_key: Optional key name to be used in the graph. By default
                  this is set to the source key converted to snake case.
        optional: Whether this key is optional. Defaults to False.
        allow_scalar: Whether this field can sometimes contain a scalar rather
                      than a list - if this is set to True the scalar will be treated as a
                      list of one. Defaults to False.
    """

    def __init__(
        self,
        source_key: str,
        sub_field: SubField,
        alti_key: str = None,
        optional: bool = False,
        allow_scalar: bool = False,
    ):
        self.source_key = source_key
        self.sub_field = sub_field
        self.alti_key = alti_key if alti_key else camel_case_to_snake_case(self.source_key)
        self.optional = optional
        self.allow_scalar = allow_scalar

    def parse(self, data: Dict[str, Any], context: Dict[str, Any]) -> List[Link]:
        """Parse this field and return a list of Links.

       Args:
           data: dictionary of data to parse
           context: context dict containing data from higher level parsing code.

        Returns:
            List of Link objects.

        Raises:
            ListFieldSourceKeyNotFoundException if self.source_key is not in data.
            ListFieldValueNotAListException if the data does not appear to represent a list.
        """
        if self.source_key not in data:
            if self.optional:
                return []
            raise ListFieldSourceKeyNotFoundException(
                f"Expected key '{self.source_key}' in data, present keys: {', '.join(data.keys())}"
            )
        sub_datas = data.get(self.source_key, [])
        if not isinstance(sub_datas, list):
            if self.allow_scalar and isinstance(sub_datas, SCALAR_TYPES):
                sub_datas = [sub_datas]
            else:
                raise ListFieldValueNotAListException(
                    (
                        f"Key '{self.source_key}' value had unexpected type, value: {sub_datas} "
                        f"type: {type(sub_datas)}"
                    )
                )
        links: List[Link] = []
        updated_context = deepcopy(context)
        updated_context.update({"parent_alti_key": self.alti_key})
        for sub_data in sub_datas:
            sub_links = self.sub_field.parse(sub_data, updated_context)
            links += sub_links
        return links


class AnonymousListField(Field):
    """An AnonymousListField is a ListField where the source_key of the field is discarded
    and not used as a name in the resulting graph. See Examples below for more clarity.

    Args:
        source_key: Name of the key in the input JSON
        field: Field object representing the type that is contained in this list.
        optional: Whether this key is optional. Defaults to False.
        allow_scalar: Whether this field can sometimes contain a scalar rather than a list - if
            this is set to True the scalar will be treated as a list of one. Defaults to False.

    Examples:
        A DictField containing an AnonymousListField:
            >>> from altimeter.core.graph.field.dict_field import DictField
            >>> from altimeter.core.graph.field.scalar_field import EmbeddedScalarField
            >>> input = {"Biota": {"Animals": ["cow", "pig", "human"],\
                                   "Plants": ["tree", "fern"]}}
            >>> field = DictField("Biota", AnonymousListField("Animals", EmbeddedScalarField()))
            >>> links = field.parse(data=input, context={})
            >>> for link in links: print(link.to_dict())
            {'pred': 'biota', 'obj': [{'pred': 'biota', 'obj': 'cow', 'type': 'simple'}, {'pred': 'biota', 'obj': 'pig', 'type': 'simple'}, {'pred': 'biota', 'obj': 'human', 'type': 'simple'}], 'type': 'multi'}
    """

    def __init__(
        self, source_key: str, field: Field, optional: bool = False, allow_scalar: bool = False
    ):
        self.source_key = source_key
        self.field = field
        self.optional = optional
        self.allow_scalar = allow_scalar

    def parse(self, data: Dict[str, Any], context: Dict[str, Any]) -> List[Link]:
        """Parse this field and return a list of Links.

       Args:
           data: dictionary of data to parse
           context: context dict containing data from higher level parsing code.

        Returns:
            List of Link objects.

        Raises:
            ListFieldSourceKeyNotFoundException if self.source_key is not in data.
            ListFieldValueNotAListException if the data does not appear to represent a list.
        """
        if self.source_key not in data:
            if self.optional:
                return []
            raise ListFieldSourceKeyNotFoundException(
                f"Expected key '{self.source_key}' in data, present keys: {', '.join(data.keys())}"
            )
        sub_datas = data.get(self.source_key, [])
        if not isinstance(sub_datas, list):
            if self.allow_scalar and isinstance(sub_datas, SCALAR_TYPES):
                sub_datas = [sub_datas]
            else:
                raise ListFieldValueNotAListException(
                    (
                        f"Key '{self.source_key}' value had unexpected type, value: {sub_datas} "
                        f"type: {type(sub_datas)}"
                    )
                )
        links: List[Link] = []
        for sub_data in sub_datas:
            sub_links = self.field.parse(sub_data, context)
            links += sub_links
        return links



"""Resource for GuardDuty Detectors"""
from typing import Any, Dict, List, Type

from botocore.client import BaseClient

from altimeter.aws.resource.guardduty import GuardDutyResourceSpec
from altimeter.aws.resource.resource_spec import ListFromAWSResult
from altimeter.core.graph.field.resource_link_field import TransientResourceLinkField
from altimeter.core.graph.field.dict_field import AnonymousDictField, EmbeddedDictField
from altimeter.core.graph.field.list_field import ListField
from altimeter.core.graph.field.scalar_field import ScalarField
from altimeter.core.graph.schema import Schema


class DetectorResourceSpec(GuardDutyResourceSpec):
    """Resource for GuardDuty Detectors"""

    type_name = "detector"
    schema = Schema(
        ScalarField("CreatedAt"),
        ScalarField("FindingPublishingFrequency"),
        ScalarField("ServiceRole"),
        ScalarField("Status"),
        ScalarField("UpdatedAt"),
        ListField(
            "Members",
            EmbeddedDictField(
                TransientResourceLinkField("DetectorArn", "DetectorResourceSpec", value_is_id=True),
                ScalarField("Email"),
                ScalarField("RelationshipStatus"),
                ScalarField("InvitedAt"),
                ScalarField("UpdatedAt"),
            ),
            alti_key="member",
        ),
        AnonymousDictField(
            "Master",
            ScalarField("AccountId", alti_key="master_account_id"),
            ScalarField("RelationshipStatus", alti_key="master_relationship_status"),
            ScalarField("InvitedAt", alti_key="master_invited_at"),
            optional=True,
        ),
    )

    @classmethod
    def list_from_aws(
        cls: Type["DetectorResourceSpec"], client: BaseClient, account_id: str, region: str
    ) -> ListFromAWSResult:
        """Return a dict of dicts of the format:

            {'detector_1_arn': {detector_1_dict},
             'detector_2_arn': {detector_2_dict},
             ...}

        Where the dicts represent results from list_detectors and list_members, get_detector for
        each listed detector."""
        list_detectors_paginator = client.get_paginator("list_detectors")
        detectors = {}
        for list_detectors_resp in list_detectors_paginator.paginate():
            detector_ids = list_detectors_resp["DetectorIds"]
            for detector_id in detector_ids:
                detector_resp = client.get_detector(DetectorId=detector_id)
                detector = {
                    key: detector_resp[key]
                    for key in (
                        "CreatedAt",
                        "FindingPublishingFrequency",
                        "ServiceRole",
                        "Status",
                        "UpdatedAt",
                    )
                }
                list_members_paginator = client.get_paginator("list_members")
                member_resps: List[Dict[str, Any]] = []
                for list_members_resp in list_members_paginator.paginate(DetectorId=detector_id):
                    member_resps += list_members_resp.get("Members", [])
                members = []
                if member_resps:
                    for member_resp in member_resps:
                        member_account_id = member_resp["AccountId"]
                        member_detector_id = member_resp["DetectorId"]
                        member_email = member_resp["Email"]
                        member_relationship_status = member_resp["RelationshipStatus"]
                        member_invited_at = member_resp["InvitedAt"]
                        member_updated_at = member_resp["UpdatedAt"]
                        member_detector_arn = cls.generate_arn(
                            account_id=member_account_id,
                            region=region,
                            resource_id=member_detector_id,
                        )
                        member = {
                            "DetectorArn": member_detector_arn,
                            "Email": member_email,
                            "RelationshipStatus": member_relationship_status,
                            "InvitedAt": member_invited_at,
                            "UpdatedAt": member_updated_at,
                        }
                        members.append(member)
                detector["Members"] = members
                master_account_resp = client.get_master_account(DetectorId=detector_id)
                master_account_dict = master_account_resp.get("Master")
                if master_account_dict:
                    detector["Master"] = {
                        key: master_account_dict[key]
                        for key in ("AccountId", "RelationshipStatus", "InvitedAt",)
                    }
                resource_arn = cls.generate_arn(
                    account_id=account_id, region=region, resource_id=detector_id
                )
                detectors[resource_arn] = detector
        return ListFromAWSResult(resources=detectors)



"""Classes for caching AWS credentials"""
from dataclasses import asdict, dataclass, field
import time
from typing import Any, Dict, Optional, Type

import boto3


@dataclass(frozen=True)
class AWSCredentials:
    """Represents a set of AWS Credentials

    Args:
        access_key_id: AWS access key id
        secret_access_key: AWS secret access key
        session_token: AWS session token
        expiration: Session expiration as an epoch timestamp int
    """

    access_key_id: str
    secret_access_key: str
    session_token: str
    expiration: int

    def is_expired(self) -> bool:
        """Determine if this cache value is within 60 seconds of expiry

        Returns:
            True if this session is value is expired, else False.
        """
        return int(time.time()) >= self.expiration - 60

    def get_session(self, region_name: Optional[str] = None) -> boto3.Session:
        """Build a boto3.Session using these credentials"""
        return boto3.Session(
            aws_access_key_id=self.access_key_id,
            aws_secret_access_key=self.secret_access_key,
            aws_session_token=self.session_token,
            region_name=region_name,
        )

    @classmethod
    def from_dict(cls: Type["AWSCredentials"], data: Dict[str, Any]) -> "AWSCredentials":
        """Build an AWSCredentials object from a dictionary"""
        return cls(**data)


@dataclass(frozen=True)
class AWSCredentialsCacheKey:
    """Represents a credential cache key

    Args:
        account_id: session account id
        role_name: session role name
        role_session_name: session role session name
    """

    account_id: str
    role_name: str
    role_session_name: str

    def __str__(self) -> str:
        return ":".join((self.account_id, self.role_name, self.role_session_name))

    @classmethod
    def from_str(cls: Type["AWSCredentialsCacheKey"], key: str) -> "AWSCredentialsCacheKey":
        account_id, role_name, role_session_name = key.split(":")
        return cls(account_id=account_id, role_name=role_name, role_session_name=role_session_name)


@dataclass(frozen=True)
class AWSCredentialsCache:
    """An AWSCredentialsCache is a cache for AWSCredentials."""

    # https://github.com/PyCQA/pylint/issues/2698
    # pylint: disable=unsupported-assignment-operation,no-member,unsupported-delete-operation,unsubscriptable-object

    cache: Dict[AWSCredentialsCacheKey, AWSCredentials] = field(default_factory=dict)

    def put(
        self, credentials: AWSCredentials, account_id: str, role_name: str, role_session_name: str
    ) -> None:
        """Put an AWSCredentials object into the cache.

        Args:
            credentials: credentials to cache
            account_id: session account id
            role_name: session role name
            role_session_name: session role session name
        """
        session = credentials.get_session()
        sts_client = session.client("sts")
        caller_id = sts_client.get_caller_identity()
        # make sure these creds are for the account the caller claims them to be
        if caller_id["Account"] != account_id:
            raise ValueError(
                (
                    f"Credentials {credentials} are not for claimed account_id "
                    f"{account_id} - they are for {caller_id['Account']}"
                )
            )
        cache_key = AWSCredentialsCacheKey(account_id, role_name, role_session_name)
        self.cache[cache_key] = credentials

    def get(
        self,
        account_id: str,
        role_name: str,
        role_session_name: str,
        region_name: Optional[str] = None,
    ) -> Optional[boto3.Session]:
        """Get a boto3 Session from AWSCredentials in the cache. Return None if no matching
        AWSCredentials were found.

        Args:
            account_id: session account id
            role_name: session role name
            role_session_name: session role session name
            region_name: session region_name

        Returns:
            boto3.Session from credentials if cached, else None.
        """
        cache_key = AWSCredentialsCacheKey(account_id, role_name, role_session_name)
        cache_val = self.cache.get(cache_key)
        if cache_val is not None:
            if cache_val.is_expired():
                del self.cache[cache_key]
            else:
                return self.cache[cache_key].get_session(region_name=region_name)
        return None

    def to_dict(self) -> Dict[str, Any]:
        return {"cache": {str(key): asdict(val) for key, val in self.cache.items()}}

    @classmethod
    def from_dict(cls: Type["AWSCredentialsCache"], data: Dict[str, Any]) -> "AWSCredentialsCache":
        cache = {
            AWSCredentialsCacheKey.from_str(key=key): AWSCredentials.from_dict(data=val)
            for key, val in data["cache"].items()
        }
        return cls(cache=cache)



"""A GraphSet represents the contents of a Graph."""
from collections import defaultdict
from typing import Any, DefaultDict, Dict, List, Type

from rdflib import BNode, Graph, Literal, Namespace, RDF

from altimeter.core.graph.exceptions import (
    GraphSetOrphanedReferencesException,
    UnmergableGraphSetsException,
)
from altimeter.core.graph.link.links import ResourceLinkLink
from altimeter.core.graph.node_cache import NodeCache
from altimeter.core.multilevel_counter import MultilevelCounter
from altimeter.core.resource.resource import Resource
from altimeter.core.resource.resource_spec import ResourceSpec


class GraphSet:
    """A GraphSet represents the contents of a Graph.  It contains a list of Resource objects, a
    graph name and version and a few metadata fields describing the scan. Generally GraphSets
    are created by running `GraphSpec.scan`.

    Args:
        name: graph name
        version: graph version
        start_time: epoch scan start time
        end_time: epoch scan end time
        resources: Resource objects
        errors: Errors encountered during scan
        stats: Scan statistics, generally API call stats.
    """

    def __init__(
        self,
        name: str,
        version: str,
        start_time: int,
        end_time: int,
        resources: List[Resource],
        errors: List[str],
        stats: MultilevelCounter,
    ):
        self.name = name
        self.version = version
        self.start_time = start_time
        self.end_time = end_time
        self.resources = resources
        self.errors = errors
        self.stats = stats
        self._resolve_duplicates()

    def to_rdf(self) -> Graph:
        """Generate an rdflib.Graph from this GraphSet.

        Returns:
            rdf.Graph object representing this GraphSet.
        """
        namespace = Namespace(f"{self.name}:")
        node_cache = NodeCache()
        graph = Graph()
        metadata_node = BNode()
        graph.add((metadata_node, RDF.type, getattr(namespace, "metadata")))
        graph.add((metadata_node, getattr(namespace, "name"), Literal(self.name)))
        graph.add((metadata_node, getattr(namespace, "version"), Literal(self.version)))
        graph.add((metadata_node, getattr(namespace, "start_time"), Literal(self.start_time)))
        graph.add((metadata_node, getattr(namespace, "end_time"), Literal(self.end_time)))
        for error in self.errors:
            graph.add((metadata_node, getattr(namespace, "error"), Literal(error)))
        for resource in self.resources:
            resource.to_rdf(namespace=namespace, graph=graph, node_cache=node_cache)
        return graph

    def to_dict(self) -> Dict[str, Any]:
        """Generate a dictionary representation of this GraphSet.

        Returns:
            dict representation of this GraphSet
        """
        resources = {resource.resource_id: resource.to_dict() for resource in self.resources}
        return {
            "name": self.name,
            "version": self.version,
            "start_time": self.start_time,
            "end_time": self.end_time,
            "resources": resources,
            "errors": self.errors,
            "stats": self.stats.to_dict(),
        }

    def validate(self) -> None:
        """Validate that all inter-resource relationships in this GraphSet resolve.

        Raises:
            GraphSetOrphanedReferencesException if references to Resources are found in
                the graph which do not exist. For example, if an EC2 instance has a VPC and
                references it as a ResourceLinkField if that VPC is not found as a Resource in the
                graph this exception is raised.
        """
        present_resource_ids = {resource.resource_id for resource in self.resources}
        resource_ref_ids_used_by_ids: DefaultDict[str, List[str]] = defaultdict(list)
        for resource in self.resources:
            for link in resource.links:
                if isinstance(link, ResourceLinkLink):
                    resource_ref_ids_used_by_ids[link.obj].append(resource.resource_id)
        resource_ref_ids = set(resource_ref_ids_used_by_ids.keys())
        orphan_refs = resource_ref_ids - present_resource_ids
        if orphan_refs:
            raise GraphSetOrphanedReferencesException(
                ("References to resources were found which were not scanned: " f"{orphan_refs}.")
            )

    def _resolve_duplicates(self) -> None:
        """Resolve any duplicate resource ids.  In general duplicate resource ids can
        have their Resource objects merged if they are of the same type and all fields
        are identical or additive only across the resources or if one of the Resources
        allows a special merge via its ResourceSpec class' `allow_clobber` attribute."""
        resource_ids_resources: DefaultDict[str, List[Resource]] = defaultdict(list)
        for resource in self.resources:
            resource_ids_resources[resource.resource_id].append(resource)
        merged_resources: List[Resource] = []
        for resource_id, resources in resource_ids_resources.items():
            if len(resources) > 1:
                merged_resource = ResourceSpec.merge_resources(
                    resource_id=resource_id, resources=resources
                )
                merged_resources.append(merged_resource)
        for merged_resource in merged_resources:
            self.resources = [
                resource
                for resource in self.resources
                if resource.resource_id != merged_resource.resource_id
            ]
            self.resources.append(merged_resource)

    @classmethod
    def from_dict(cls: Type["GraphSet"], data: Dict[str, Any]) -> "GraphSet":
        """Create a GraphSet from a dict.

        Args:
            data: dict of Resource data

        Returns:
            GraphSet object
        """
        resources: List[Resource] = []
        name = data["name"]
        start_time = data["start_time"]
        end_time = data["end_time"]
        version = data["version"]
        errors = data["errors"]
        stats = MultilevelCounter.from_dict(data["stats"])
        for resource_id, resource_data in data["resources"].items():
            resource = Resource.from_dict(resource_id, resource_data)
            resources.append(resource)
        return cls(
            name=name,
            version=version,
            start_time=start_time,
            end_time=end_time,
            resources=resources,
            errors=errors,
            stats=stats,
        )

    def merge(self, other: "GraphSet") -> None:
        """Merge another GraphSet into this GraphSet.

        Args:
            other: GraphSet to merge into this GraphSet.
        """
        if other.name != self.name:
            raise UnmergableGraphSetsException(
                f"Unable to merge graph with name {other.name} into {self.name}"
            )
        if other.version != self.version:
            raise UnmergableGraphSetsException(
                f"Unable to merge graph with version {other.version} into {self.version}"
            )
        self.start_time = min(self.start_time, other.start_time)
        self.end_time = max(self.end_time, other.end_time)
        self.resources += other.resources
        self._resolve_duplicates()
        self.errors += other.errors
        self.stats.merge(other.stats)



"""Scalar Fields represent field that contain strings, bools or numbers."""
import datetime
from typing import Union, Dict, Any, List

from altimeter.core.graph.field.exceptions import (
    ScalarFieldSourceKeyNotFoundException,
    ScalarFieldValueNotAScalarException,
)
from altimeter.core.graph.field.base import Field, SubField
from altimeter.core.graph.field.util import camel_case_to_snake_case
from altimeter.core.graph.link.links import SimpleLink
from altimeter.core.graph.link.base import Link

SCALAR_TYPES = (str, bool, int, float, datetime.datetime)


class ScalarField(Field):
    """A ScalarField is a field where the input is a JSON object and the corresponding value is a
    string, number or bool.

    Examples:
        A ScalarField with a string value::
            >>> input = {"FieldName": "Value"}
            >>> field = ScalarField("FieldName")
            >>> links = field.parse(data=input, context={})
            >>> print([link.to_dict() for link in links])
            [{'pred': 'field_name', 'obj': 'Value', 'type': 'simple'}]

        A ScalarField with a string value and an alti_key specified::
            >>> input = {"FieldName": "Value"}
            >>> field = ScalarField("FieldName", alti_key="custom_name")
            >>> links = field.parse(data=input, context={})
            >>> print([link.to_dict() for link in links])
            [{'pred': 'custom_name', 'obj': 'Value', 'type': 'simple'}]

        An optional ScalarField with no default value::
            >>> input = {"SomeOtherFieldName": "Value"}
            >>> field = ScalarField("FieldName", optional=True)
            >>> links = field.parse(data=input, context={})
            >>> print([link.to_dict() for link in links])
            []

        An optional ScalarField with a default value::
            >>> input = {"SomeOtherFieldName": "Value"}
            >>> field = ScalarField("FieldName", optional=True, default_value="DefaultValue")
            >>> links = field.parse(data=input, context={})
            >>> print([link.to_dict() for link in links])
            [{'pred': 'field_name', 'obj': 'DefaultValue', 'type': 'simple'}]

    Args:
        source_key: Name of the key in the input JSON
        alti_key: Optional key name to be used in the graph. By default this is set to the source
                  key converted to snake case.
        optional: Whether this key is optional. Defaults to False.
        default_value: A default value to use if the source_key is not present.
    """

    def __init__(
        self,
        source_key: str,
        alti_key: str = None,
        optional: bool = False,
        default_value: Union[str, bool, int, float] = None,
    ):
        self.source_key = source_key
        self.alti_key = alti_key if alti_key else camel_case_to_snake_case(self.source_key)
        self.optional = optional
        self.default_value = default_value

    def parse(self, data: Dict[str, Any], context: Dict[str, Any]) -> List[Link]:
        """Parse this field and return a list of Links.

       Args:
           data: dictionary of data to parse
           context: context dict containing data from higher level parsing code.

        Returns:
            List of SimpleLink objects. At most one link will be returned.
        """
        value = data.get(self.source_key)
        if value is None:
            if self.default_value is not None:
                return [SimpleLink(pred=self.alti_key, obj=self.default_value)]
            if self.optional:
                return []
            raise ScalarFieldSourceKeyNotFoundException(
                f"Expected key '{self.source_key}' in data, present keys: {', '.join(data.keys())}"
            )
        if isinstance(value, SCALAR_TYPES):
            return [SimpleLink(pred=self.alti_key, obj=value)]
        raise ScalarFieldValueNotAScalarException(
            (
                f"Expected data for key '{self.source_key}' to be one "
                f"of {SCALAR_TYPES}, is {type(value)}: {value}"
            )
        )


class EmbeddedScalarField(SubField):
    """An EmbeddedScalarField is a field where the input is a string, number or bool.
    An EmbeddedScalarField assumes the key of the enclosing field.

    Examples:
        An EmbeddedScalarField inside a ListField::
            >>> from altimeter.core.graph.field.list_field import ListField
            >>> input = {"Animal": ["Value1", "Value2"]}
            >>> field = ListField("Animal", EmbeddedScalarField())
            >>> links = field.parse(data=input, context={})
            >>> for link in links:
            ...     print(link.to_dict())
            ...
            {'pred': 'animal', 'obj': 'Value1', 'type': 'simple'}
            {'pred': 'animal', 'obj': 'Value2', 'type': 'simple'}
    """

    def parse(self, data: Union[str, bool, int, float], context: Dict[str, Any]) -> List[Link]:
        """Parse this field and return a list of Links. For a ScalarField at most one link
        will be returned.

        Args:
            data: scalar value
            context: context dict containing data from higher level parsing code.

        Returns:
            List of Link objects. At most one link will be returned.
        """
        parent_alti_key = self.get_parent_alti_key(data, context)
        links: List[Link] = []
        if isinstance(data, SCALAR_TYPES):
            link = SimpleLink(pred=parent_alti_key, obj=data)
            links.append(link)
            return links
        raise ScalarFieldValueNotAScalarException(
            (f"Expected data to be one of {SCALAR_TYPES}, is " f"{type(data)}: {data}")
        )



"""Resource for VPC Flow Logs"""
from typing import Type

from botocore.client import BaseClient

from altimeter.aws.resource.resource_spec import ListFromAWSResult
from altimeter.aws.resource.ec2 import EC2ResourceSpec
from altimeter.aws.resource.ec2.vpc import VPCResourceSpec
from altimeter.core.graph.field.resource_link_field import TransientResourceLinkField
from altimeter.core.graph.field.scalar_field import ScalarField
from altimeter.core.graph.schema import Schema


class FlowLogResourceSpec(EC2ResourceSpec):
    """Resource for VPC Flow Logs"""

    type_name = "flow-log"
    schema = Schema(
        ScalarField("CreationTime"),
        ScalarField("DeliverLogsErrorMessage", optional=True),
        ScalarField("DeliverLogsPermissionArn", optional=True),
        ScalarField("DeliverLogsStatus", optional=True),
        ScalarField("FlowLogStatus"),
        ScalarField("LogGroupName", optional=True),
        TransientResourceLinkField("ResourceId", VPCResourceSpec, optional=True),
        ScalarField("TrafficType"),
        ScalarField("LogDestinationType"),
        ScalarField("LogDestination", optional=True),
        ScalarField("LogFormat"),
    )

    @classmethod
    def list_from_aws(
        cls: Type["FlowLogResourceSpec"], client: BaseClient, account_id: str, region: str
    ) -> ListFromAWSResult:
        """Return a dict of dicts of the format:

            {'fl_1_arn': {fl_1_dict},
             'fl_2_arn': {fl_2_dict},
             ...}

        Where the dicts represent results from describe_flow_logs."""
        flow_logs = {}
        paginator = client.get_paginator("describe_flow_logs")
        for resp in paginator.paginate():
            for flow_log in resp.get("FlowLogs", []):
                resource_arn = cls.generate_arn(
                    account_id=account_id, region=region, resource_id=flow_log["FlowLogId"]
                )
                flow_logs[resource_arn] = flow_log
        return ListFromAWSResult(resources=flow_logs)



"""An AccountScanManifest defines the output of an account scan."""

from dataclasses import dataclass
from typing import Any, Dict, List


@dataclass(frozen=True)
class AccountScanManifest:
    """An AccountScanManifest defines the output of an account scan. It contains pointers to the
    scan result artifacts and summaries of what was scanned, errors which occurred, and api call
    statistics.

    Args:
        account_id: account id
        artifacts: list of scan artifacts
        errors: list of error strings
        api_call_stats: dict of api call stats
    """

    account_id: str
    artifacts: List[str]
    errors: List[str]
    api_call_stats: Dict[str, Any]



"""Resource for RDS"""
from typing import Type

from botocore.client import BaseClient

from altimeter.aws.log import AWSLogEvents
from altimeter.aws.resource.resource_spec import ListFromAWSResult
from altimeter.aws.resource.ec2.security_group import SecurityGroupResourceSpec
from altimeter.aws.resource.ec2.vpc import VPCResourceSpec
from altimeter.aws.resource.rds import RDSResourceSpec
from altimeter.aws.resource.kms.key import KMSKeyResourceSpec
from altimeter.core.graph.field.dict_field import (
    AnonymousDictField,
    AnonymousEmbeddedDictField,
    EmbeddedDictField,
)
from altimeter.core.graph.field.list_field import AnonymousListField, ListField
from altimeter.core.graph.field.resource_link_field import TransientResourceLinkField
from altimeter.core.graph.field.scalar_field import ScalarField
from altimeter.core.graph.field.tags_field import TagsField
from altimeter.core.graph.schema import Schema
from altimeter.core.log import Logger


class RDSInstanceResourceSpec(RDSResourceSpec):
    """Resource for RDS"""

    type_name = "db"
    schema = Schema(
        TagsField(),
        ScalarField("DBInstanceIdentifier"),
        ScalarField("DBInstanceClass"),
        ScalarField("Engine"),
        ScalarField("DBInstanceStatus"),
        ScalarField("DBName", optional=True),
        AnonymousDictField(
            "Endpoint",
            ScalarField("Address", alti_key="endpoint_address", optional=True),
            ScalarField("Port", alti_key="endpoint_port"),
            ScalarField("HostedZoneId", alti_key="endpoint_hosted_zone", optional=True),
            optional=True,
        ),
        AnonymousDictField(
            "ListenerEndpoint",
            ScalarField("Address", alti_key="listener_address"),
            ScalarField("Port", alti_key="listener_port"),
            ScalarField("HostedZoneId", alti_key="listener_hosted_zone", optional=True),
            optional=True,
        ),
        ScalarField("InstanceCreateTime", optional=True),
        ScalarField("BackupRetentionPeriod"),
        AnonymousListField(
            "VpcSecurityGroups",
            AnonymousEmbeddedDictField(
                TransientResourceLinkField(
                    "VpcSecurityGroupId", SecurityGroupResourceSpec, optional=True
                )
            ),
        ),
        ScalarField("AvailabilityZone", optional=True),
        AnonymousDictField(
            "DBSubnetGroup", TransientResourceLinkField("VpcId", VPCResourceSpec), optional=True
        ),
        ScalarField("MultiAZ"),
        ScalarField("PubliclyAccessible"),
        ListField("StatusInfos", EmbeddedDictField(ScalarField("Status")), optional=True),
        ScalarField("StorageType"),
        ScalarField("StorageEncrypted"),
        TransientResourceLinkField("KmsKeyId", KMSKeyResourceSpec, optional=True, value_is_id=True),
        ScalarField("DbiResourceId"),
        ScalarField("Timezone", optional=True),
        ScalarField("IAMDatabaseAuthenticationEnabled"),
        ScalarField("PerformanceInsightsEnabled", optional=True),
        ScalarField("PerformanceInsightsRetentionPeriod", optional=True),
        ScalarField("DeletionProtection"),
        ListField(
            "Backup",
            EmbeddedDictField(
                AnonymousDictField(
                    "RestoreWindow",
                    ScalarField("EarliestTime", alti_key="earliest_restore_time", optional=True),
                    optional=True,
                ),
                AnonymousDictField(
                    "RestoreWindow",
                    ScalarField("LatestTime", alti_key="latest_restore_time", optional=True),
                    optional=True,
                ),
                ScalarField("AllocatedStorage"),
                ScalarField("Status"),
                ScalarField("AvailabilityZone", optional=True),
                ScalarField("Engine"),
                ScalarField("EngineVersion"),
                ScalarField("Encrypted"),
                ScalarField("StorageType"),
                TransientResourceLinkField(
                    "KmsKeyId", KMSKeyResourceSpec, optional=True, value_is_id=True
                ),
            ),
            optional=True,
        ),
    )

    @classmethod
    def list_from_aws(
        cls: Type["RDSInstanceResourceSpec"], client: BaseClient, account_id: str, region: str
    ) -> ListFromAWSResult:
        logger = Logger()
        dbinstances = {}
        paginator = client.get_paginator("describe_db_instances")
        for resp in paginator.paginate():
            for db in resp.get("DBInstances", []):
                resource_arn = db["DBInstanceArn"]
                db["Tags"] = client.list_tags_for_resource(ResourceName=resource_arn).get(
                    "TagList", []
                )
                db["Backup"] = []
                dbinstances[resource_arn] = db

        backup_paginator = client.get_paginator("describe_db_instance_automated_backups")
        for resp in backup_paginator.paginate():
            for backup in resp.get("DBInstanceAutomatedBackups", []):
                if backup["DBInstanceArn"] in dbinstances:
                    dbinstances[backup["DBInstanceArn"]]["Backup"].append(backup)
                else:
                    logger.info(
                        event=AWSLogEvents.ScanAWSResourcesNonFatalError,
                        msg=(
                            f'Unable to find matching DB Instance {backup["DBInstanceArn"]} '
                            "(Possible Deletion)"
                        ),
                    )
        return ListFromAWSResult(resources=dbinstances)



"""Base class for AWS organizations resources."""
from typing import Type

from botocore.client import BaseClient

from altimeter.aws.resource.resource_spec import ScanGranularity, AWSResourceSpec


class OrganizationsResourceSpec(AWSResourceSpec):
    """Base class for AWS organizations resources."""

    service_name = "organizations"
    scan_granularity = ScanGranularity.ACCOUNT

    @classmethod
    def skip_resource_scan(
        cls: Type["OrganizationsResourceSpec"], client: BaseClient, account_id: str, region: str
    ) -> bool:
        """Return a bool indicating whether this resource class scan should be skipped,
        in this case skip if the current account is not an org master."""
        resp = client.describe_organization()
        return resp["Organization"]["MasterAccountId"] != account_id



"""Dict Fields represent fields which consist of dict-like data."""
from copy import deepcopy
from typing import Dict, Any, List

from altimeter.core.graph.field.exceptions import (
    DictFieldValueNotADictException,
    DictFieldSourceKeyNotFoundException,
    DictFieldValueIsNullException,
)
from altimeter.core.graph.field.base import Field, SubField
from altimeter.core.graph.field.util import camel_case_to_snake_case
from altimeter.core.graph.link.links import MultiLink
from altimeter.core.graph.link.base import Link


class DictField(Field):
    """A DictField is a field where the input is a JSON object containing a key (source_key)
    where the corresponding value is a dictionary.

    Examples:
        A dictionary containing two ScalarFields:
            >>> from altimeter.core.graph.field.scalar_field import ScalarField
            >>> input = {"Person": {"FirstName": "Bob", "LastName": "Smith"}}
            >>> field = DictField("Person", ScalarField("FirstName"), ScalarField("LastName"))
            >>> links = field.parse(data=input, context={})
            >>> print([link.to_dict() for link in links])
            [{'pred': 'person', 'obj': [{'pred': 'first_name', 'obj': 'Bob', 'type': 'simple'}, {'pred': 'last_name', 'obj': 'Smith', 'type': 'simple'}], 'type': 'multi'}]

    Args:
        source_key: Name of the key in the input JSON
        fields: fields inside this DictField
        alti_key: Optional key name to be used in the graph. By default
                  this is set to the source key converted to snake case.
        optional: Whether this key is optional. Defaults to False.
    """

    def __init__(
        self, source_key: str, *fields: Field, alti_key: str = None, optional: bool = False
    ) -> None:
        self.source_key = source_key
        self.alti_key = alti_key if alti_key else camel_case_to_snake_case(self.source_key)
        self.optional = optional
        self.fields = fields

    def parse(self, data: Dict[str, Any], context: Dict[str, Any]) -> List[Link]:
        """Parse this field and return a list of Links.

       Args:
           data: dictionary of data to parse
           context: context dict containing data from higher level parsing code.

        Returns:
            List of MultiLink objects.

        Raises:
            DictFieldSourceKeyNotFoundException if self.source_key is not in data.
            DictFieldValueNotADictException if the data does not appear to represent a dict.
        """
        if self.source_key not in data:
            if self.optional:
                return []
            raise DictFieldSourceKeyNotFoundException(
                f"Expected key '{self.source_key}' in data, present keys: {', '.join(data.keys())}"
            )
        field_data = data.get(self.source_key, {})
        if not isinstance(field_data, dict):
            raise DictFieldValueNotADictException(
                (
                    f"Key '{self.source_key}' value was expected to "
                    f"contain a dict, actual: {field_data} "
                    f"({type(field_data)})"
                )
            )
        links: List[Link] = []
        updated_context = deepcopy(context)
        updated_context.update({"parent_alti_key": self.alti_key})
        for field in self.fields:
            sub_links = field.parse(field_data, updated_context)
            links += sub_links
        return [MultiLink(pred=self.alti_key, obj=links)]


class AnonymousDictField(Field):
    """An AnonymousDictField is a DictField where the source_key of the field is discarded
    and not used as a name in the resulting graph. See Examples below for more clarity.

    Args:
        source_key: Name of the key in the input JSON
        fields: fields inside this DictField
        optional: Whether this key is optional. Defaults to False.
        nullable: Whether this field's value can be null.

    Examples:
        A dict containing 3 ScalarFields
            >>> from altimeter.core.graph.field.scalar_field import ScalarField
            >>> input = {"Person": {"FirstName": "Bob", "LastName": "Smith"}}
            >>> field = AnonymousDictField("Person", ScalarField("FirstName"), ScalarField("LastName"))
            >>> links = field.parse(data=input, context={})
            >>> for link in links: print(link.to_dict())
            {'pred': 'first_name', 'obj': 'Bob', 'type': 'simple'}
            {'pred': 'last_name', 'obj': 'Smith', 'type': 'simple'}
    """

    def __init__(
        self, source_key: str, *fields: Field, optional: bool = False, nullable: bool = False
    ):
        self.source_key = source_key
        self.fields = fields
        self.optional = optional
        self.nullable = nullable

    def parse(self, data: Dict[str, Any], context: Dict[str, Any]) -> List[Link]:
        """Parse this field and return a list of Links.

       Args:
           data: dictionary of data to parse
           context: context dict containing data from higher level parsing code.

        Returns:
            List of Link objects.

        Raises:
            DictFieldSourceKeyNotFoundException if self.source_key is not in data.
            DictFieldValueNotADictException if the data does not appear to represent a dict.
        """
        if self.source_key in data:
            field_data = data.get(self.source_key, None)
            if field_data is None:
                if self.nullable:
                    return []
                raise DictFieldValueIsNullException(
                    f"Key '{self.source_key} was expected to contain a dict, was null."
                )
            if not isinstance(field_data, dict):
                raise DictFieldValueNotADictException(
                    (
                        f"Key '{self.source_key}' value expected to "
                        f"contain a dict, actual: {field_data} "
                        f"({type(field_data)})"
                    )
                )
            links: List[Link] = []
            for field in self.fields:
                sub_links = field.parse(field_data, context)
                links += sub_links
            return links
        if self.optional:
            return []
        raise DictFieldSourceKeyNotFoundException(
            f"Expected key '{self.source_key}' in data, present keys: {', '.join(data.keys())}"
        )


class EmbeddedDictField(SubField):
    """An EmbeddedDictField is a field where the input is a JSON object.  Generally this field
    is used inside a ListField.

    Args:
        fields: fields inside this DictField

    Examples:
        A ListField containing an EmbeddedDictField with two ScalarFields:
            >>> from altimeter.core.graph.field.list_field import ListField
            >>> from altimeter.core.graph.field.scalar_field import ScalarField
            >>> input = {"People": [{"FirstName": "Bob", "LastName": "Smith"},\
                         {"FirstName": "Alice", "LastName": "Smith"}]}
            >>> field = ListField("People", EmbeddedDictField(ScalarField("FirstName"),\
                                  ScalarField("LastName")))
            >>> links = field.parse(data=input, context={})
            >>> for link in links: print(link.to_dict())
            {'pred': 'people', 'obj': [{'pred': 'first_name', 'obj': 'Bob', 'type': 'simple'}, {'pred': 'last_name', 'obj': 'Smith', 'type': 'simple'}], 'type': 'multi'}
            {'pred': 'people', 'obj': [{'pred': 'first_name', 'obj': 'Alice', 'type': 'simple'}, {'pred': 'last_name', 'obj': 'Smith', 'type': 'simple'}], 'type': 'multi'}
    """

    def __init__(self, *fields: Field) -> None:
        self.fields = fields

    def parse(self, data: Dict[str, Any], context: Dict[str, Any]) -> List[Link]:
        """Parse this field and return a list of Links.

        Args:
            data: dictionary of data to parse
            context: context dict containing data from higher level parsing code.

        Returns:
            List of MultiLink objects.

        Raises:
            DictFieldSourceKeyNotFoundException if self.source_key is not in data.
            DictFieldValueNotADictException if the data does not appear to represent a dict.
        """
        parent_alti_key = self.get_parent_alti_key(data, context)
        if not isinstance(data, dict):
            raise Exception(f"{type(data)} {data} was expected to be a dict.")
        links: List[Link] = []
        updated_context = deepcopy(context)
        updated_context.update({"parent_alti_key": parent_alti_key})
        for field in self.fields:
            sub_links = field.parse(data, updated_context)
            links += sub_links
        return [MultiLink(pred=parent_alti_key, obj=links)]


class AnonymousEmbeddedDictField(Field):
    """An AnonymousEmbeddedDictField is a EmbeddedDictField where the source_key of the parent
    field is discarded and not used as a name in the resulting graph. See Examples below for more
    clarity.

    Args:
        fields: fields inside this DictField

    Examples:
        A ListField containing an AnonymousEmbeddedDictField with two ScalarFields:
            >>> from altimeter.core.graph.field.list_field import ListField
            >>> from altimeter.core.graph.field.scalar_field import ScalarField
            >>> input = {"People": [{"FirstName": "Bob", "LastName": "Smith"},\
                         {"FirstName": "Alice", "LastName": "Smith"}]}
            >>> field = ListField("People", AnonymousEmbeddedDictField(ScalarField("FirstName"),\
                                  ScalarField("LastName")))
            >>> links = field.parse(data=input, context={})
            >>> for link in links: print(link.to_dict())
            {'pred': 'first_name', 'obj': 'Bob', 'type': 'simple'}
            {'pred': 'last_name', 'obj': 'Smith', 'type': 'simple'}
            {'pred': 'first_name', 'obj': 'Alice', 'type': 'simple'}
            {'pred': 'last_name', 'obj': 'Smith', 'type': 'simple'}
    """

    def __init__(self, *fields: Field) -> None:
        self.fields = fields

    def parse(self, data: Dict[str, Any], context: Dict[str, Any]) -> List[Link]:
        """Parse this field and return a list of Links.

        Args:
            data: dictionary of data to parse
            context: context dict containing data from higher level parsing code.

        Returns:
            List of Link objects.

        Raises:
            DictFieldValueNotADictException if the data does not appear to represent a dict.
        """
        if not isinstance(data, dict):
            raise DictFieldValueNotADictException(f"{type(data)} {data} was expected to be a dict.")
        links: List[Link] = []
        for field in self.fields:
            sub_links = field.parse(data, context)
            links += sub_links
        return links



"""Resource Link Fields represent field containing ids of other top level resources in the graph.
For example, an EC2 instance has a ResourceLinkField with source_key 'VpcId' pointing to a VPC."""
from typing import Dict, Any, List, Type, Union

from altimeter.core.graph.field.exceptions import (
    ResourceLinkFieldSourceKeyNotFoundException,
    ResourceLinkFieldValueNotAStringException,
)
from altimeter.core.graph.field.base import Field, SubField
from altimeter.core.graph.link.links import ResourceLinkLink, TransientResourceLinkLink
from altimeter.core.graph.link.base import Link
from altimeter.core.resource.resource_spec import ResourceSpec


class ResourceLinkField(Field):
    """A ResourceLinkField represents a field containing ids of other top level resources in the
    graph. For example, an EC2 instance has a ResourceLinkField with source_key 'VpcId' pointing to
    a VPC.

    Examples:
        A link to a TestResourceSpec resource::
            >>> from altimeter.core.resource.resource_spec import ResourceSpec
            >>> class TestResourceSpec(ResourceSpec): type_name="thing"
            >>> input = {"ThingId": "123"}
            >>> field = ResourceLinkField("ThingId", TestResourceSpec)
            >>> links = field.parse(data=input, context={})
            >>> print([link.to_dict() for link in links])
            [{'pred': 'thing', 'obj': 'thing:123', 'type': 'resource_link'}]

        A link to a TestResourceSpec resource using value_is_id::
            >>> from altimeter.core.resource.resource_spec import ResourceSpec
            >>> class TestResourceSpec(ResourceSpec): type_name="thing"
            >>> input = {"ThingId": "thing:123"}
            >>> field = ResourceLinkField("ThingId", TestResourceSpec, value_is_id=True)
            >>> links = field.parse(data=input, context={})
            >>> print([link.to_dict() for link in links])
            [{'pred': 'thing', 'obj': 'thing:123', 'type': 'resource_link'}]

    Args:
        source_key: Name of the key in the input JSON
        resource_spec_class: The name of the ResourceSpec class or the ResourceSpec class which this
                             link represents.
        alti_key: Optional key name to be used in the graph. By default this is set to the
                  resource_spec_class' type_name attribute.
        optional: Whether this key is optional. Defaults to False.
        value_is_id: Whether the value for this key contains the entire resource id. For AWS
                     resources set this to True if the value is a complete arn.
    """

    def __init__(
        self,
        source_key: str,
        resource_spec_class: Union[Type[ResourceSpec], str],
        alti_key: str = None,
        optional: bool = False,
        value_is_id: bool = False,
    ):
        self.source_key = source_key
        self._resource_spec_class = resource_spec_class
        self.alti_key = alti_key
        self.optional = optional
        self.value_is_id = value_is_id

    def parse(self, data: Dict[str, Any], context: Dict[str, Any]) -> List[Link]:
        """Parse this field and return a list of Links.

        Args:
            data: data to parse
            context: contains data from higher level parsing code.

        Returns:
             List of Link objects. At most one link will be returned.
        """
        if isinstance(self._resource_spec_class, str):
            resource_spec_class: Type[ResourceSpec] = ResourceSpec.get_by_class_name(
                self._resource_spec_class
            )
        else:
            resource_spec_class = self._resource_spec_class
        if not self.alti_key:
            self.alti_key = resource_spec_class.type_name
        short_resource_id = data.get(self.source_key)
        if not short_resource_id:
            if self.optional:
                return []
            raise ResourceLinkFieldSourceKeyNotFoundException(
                f"Expected key '{self.source_key}' with non-empty/zero value in {data}"
            )
        if not isinstance(short_resource_id, str):
            raise ResourceLinkFieldValueNotAStringException(
                (
                    f"ResourceLinkField for {self.source_key} expected a string but got a "
                    f"{type(short_resource_id)} : {short_resource_id}"
                )
            )
        if self.value_is_id:
            resource_id = short_resource_id
        else:
            resource_id = resource_spec_class.generate_id(short_resource_id, context)
        return [ResourceLinkLink(pred=self.alti_key, obj=resource_id)]


class EmbeddedResourceLinkField(SubField):
    """An EmbeddedResourceLinkField is a ResourceLinkField where the input is the resource id
    only, not a key/value where the value is a resource id.

    Examples:
        A link to a TestResourceSpec resource::
            >>> from altimeter.core.graph.field.list_field import ListField
            >>> from altimeter.core.resource.resource_spec import ResourceSpec
            >>> class TestResourceSpec(ResourceSpec): type_name="thing"
            >>> input = {"Thing": ["123", "456"]}
            >>> field = ListField("Thing", EmbeddedResourceLinkField(TestResourceSpec))
            >>> links = field.parse(data=input, context={})
            >>> print([link.to_dict() for link in links])
            [{'pred': 'thing', 'obj': 'thing:123', 'type': 'resource_link'}, {'pred': 'thing', 'obj': 'thing:456', 'type': 'resource_link'}]

    Args:
        resource_spec_class: The name of the ResourceSpec class or the ResourceSpec class which
                             this link represents.
        optional: Whether this key is optional. Defaults to False.
        value_is_id: Whether the value for this key contains the entire resource id. For AWS
                     resources set this to True if the value is a complete arn.
    """

    def __init__(
        self,
        resource_spec_class: Union[Type[ResourceSpec], str],
        alti_key: str = None,
        optional: bool = False,
        value_is_id: bool = False,
    ):
        self._resource_spec_class = resource_spec_class
        self.alti_key = alti_key
        self.optional = optional
        self.value_is_id = value_is_id

    def parse(self, data: str, context: Dict[str, Any]) -> List[Link]:
        """Parse this field and return a list of Links.

        Args:
            data: data to parse
            context: contains data from higher level parsing code.

        Returns:
            List of Link objects. At most one link will be returned.
        """
        if isinstance(self._resource_spec_class, str):
            resource_spec_class: Type[ResourceSpec] = ResourceSpec.get_by_class_name(
                self._resource_spec_class
            )
        else:
            resource_spec_class = self._resource_spec_class
        if not self.alti_key:
            self.alti_key = resource_spec_class.type_name

        short_resource_id = data
        if self.value_is_id:
            resource_id = short_resource_id
        else:
            resource_id = resource_spec_class.generate_id(short_resource_id, context)
        return [ResourceLinkLink(pred=self.alti_key, obj=resource_id)]


class TransientResourceLinkField(Field):
    """Transient Resource Link Fields represent field containing ids of other top level resources in
    the graph which may not exist. For example, a Lambda can refer to a VPC however VPCs can be
    deleted from lambdas.


    Args:
        source_key (str): Name of the key in the input JSON
        resource_spec_class (str|Type[ResourceSpec]): The name of the ResourceSpec class or the
            ResourceSpec class which this link represents.
        alti_key (str): Optional key name to be used in the graph. By default
            this is set to the resource_spec_class' type_name attribute.
        optional (bool): Whether this key is optional. Defaults to False.
        value_is_id(bool): Whether the value for this key contains the entire resource id.
            For AWS resources set this to True if the value is a complete arn.

    Examples:
        A link to a TestResourceSpec resource::
            >>> from altimeter.core.resource.resource_spec import ResourceSpec
            >>> class TestResourceSpec(ResourceSpec): type_name="thing"
            >>> input = {"ThingId": "123"}
            >>> field = TransientResourceLinkField("ThingId", TestResourceSpec)
            >>> links = field.parse(data=input, context={})
            >>> print([link.to_dict() for link in links])
            [{'pred': 'thing', 'obj': 'thing:123', 'type': 'transient_resource_link'}]
    """

    def __init__(
        self,
        source_key: str,
        resource_spec_class: Union[Type[ResourceSpec], str],
        alti_key: str = None,
        optional: bool = False,
        value_is_id: bool = False,
    ):
        self.source_key = source_key
        self._resource_spec_class = resource_spec_class
        self.alti_key = alti_key
        self.optional = optional
        self.value_is_id = value_is_id

    def parse(self, data: Dict[str, Any], context: Dict[str, Any]) -> List[Link]:
        """Parse this field and return a list of Links.

        Args:
            data: data to parse
            context: contains data from higher level parsing code.

        Returns:
            List of Link objects. At most one link will be returned.
        """
        if isinstance(self._resource_spec_class, str):
            resource_spec_class: Type[ResourceSpec] = ResourceSpec.get_by_class_name(
                self._resource_spec_class
            )
        else:
            resource_spec_class = self._resource_spec_class
        if not self.alti_key:
            self.alti_key = resource_spec_class.type_name
        short_resource_id = data.get(self.source_key)
        if not short_resource_id:
            if self.optional:
                return []
            raise ResourceLinkFieldSourceKeyNotFoundException(
                f"Expected key '{self.source_key}' with non-empty/zero value in {data}"
            )
        if self.value_is_id:
            resource_id = short_resource_id
        else:
            resource_id = resource_spec_class.generate_id(short_resource_id, context)
        return [TransientResourceLinkLink(pred=self.alti_key, obj=resource_id)]



"""A GraphSpec contains a specification to scan and create a graph."""
from typing import Any, List, Tuple, Type

from altimeter.core.log import LogEvent, Logger

from altimeter.core.resource.resource import Resource
from altimeter.core.resource.resource_spec import ResourceSpec


class GraphSpec:
    """A GraphSpec contains a specification to scan and create a graph. It contains a set of
    ResourceSpec classes defining what to scan and a scan_accessor object defining how to scan.

    Args:
        name: graph name
        version: graph version
        resource_spec_classes: tuple of ResourceSpec classes
        scan_accessor: object which is passed to ResourceSpec.scan for each ResourceSpec in
                       resource_spec_classes. It should provide methods which ResourceSpec.scan
                       can use to access whatever API the ResourceSpec needs to access.
    """

    def __init__(
        self,
        name: str,
        version: str,
        resource_spec_classes: Tuple[Type[ResourceSpec], ...],
        scan_accessor: Any,
    ):
        self.name = name
        self.version = version
        self.resource_spec_classes = resource_spec_classes
        self.scan_accessor = scan_accessor

    def scan(self) -> List[Resource]:
        """Perform a scan on all of the resource classes in this GraphSpec and return
        a list of Resource objects.

        Returns:
            List of Resource objects
        """
        resources: List[Resource] = []
        logger = Logger()
        for resource_spec_class in self.resource_spec_classes:
            with logger.bind(resource_type=str(resource_spec_class.type_name)):
                logger.debug(event=LogEvent.ScanResourceTypeStart)
                scanned_resources = resource_spec_class.scan(scan_accessor=self.scan_accessor)
                resources += scanned_resources
                logger.debug(event=LogEvent.ScanResourceTypeEnd)
        return resources



"""LogEvent for AWS related events."""
from dataclasses import dataclass

from altimeter.core.log import BaseLogEvent, EventName


@dataclass(frozen=True)
class AWSLogEvents(BaseLogEvent):
    """AWS specific Log event names"""

    AuthToAccountStart: EventName
    AuthToAccountEnd: EventName
    AuthToAccountFailure: EventName

    GetSubAccountsStart: EventName
    GetSubAccountsEnd: EventName

    RunAccountScanLambdaStart: EventName
    RunAccountScanLambdaEnd: EventName

    MuxerQueueScan: EventName
    MuxerStart: EventName
    MuxerEnd: EventName
    MuxerStat: EventName

    ScanAWSAccountsStart: EventName
    ScanAWSAccountsEnd: EventName

    ScanAWSAccountStart: EventName
    ScanAWSAccountEnd: EventName
    ScanAWSAccountError: EventName

    ScanAWSAccountServiceStart: EventName
    ScanAWSAccountServiceEnd: EventName

    ScanAWSResourcesNonFatalError: EventName

    ScanConfigured: EventName



#!/usr/bin/env python3
"""Convert intermediate JSON to RDF."""
from dataclasses import dataclass
import io
import gzip
import json
import sys
import urllib

import boto3
from botocore.client import BaseClient
from rdflib import Graph

from altimeter.core.awslambda import get_required_lambda_env_var
from altimeter.core.graph.graph_set import GraphSet
from altimeter.core.log import Logger, LogEvent


@dataclass(frozen=True)
class GraphPackage:
    """Represents a rdflib.Graph and associated metadata."""

    graph: Graph
    name: str
    version: str
    start_time: int
    end_time: int


def graph_set_from_s3(s3_client: BaseClient, json_bucket: str, json_key: str) -> GraphSet:
    """Load a GraphSet from json located in an s3 object."""
    logger = Logger()
    logger.info(event=LogEvent.ReadFromS3Start)
    with io.BytesIO() as json_bytes_buf:
        s3_client.download_fileobj(json_bucket, json_key, json_bytes_buf)
        json_bytes_buf.flush()
        json_bytes_buf.seek(0)
        graph_set_bytes = json_bytes_buf.read()
        logger.info(event=LogEvent.ReadFromS3End)
    graph_set_str = graph_set_bytes.decode("utf-8")
    graph_set_dict = json.loads(graph_set_str)
    return GraphSet.from_dict(graph_set_dict)


def graph_pkg_from_s3(s3_client: BaseClient, json_bucket: str, json_key: str) -> GraphPackage:
    """Create an GraphPackage object from json content in an s3 object."""
    graph_set = graph_set_from_s3(s3_client=s3_client, json_bucket=json_bucket, json_key=json_key)
    return GraphPackage(
        graph=graph_set.to_rdf(),
        name=graph_set.name,
        version=graph_set.version,
        start_time=graph_set.start_time,
        end_time=graph_set.end_time,
    )


def lambda_handler(event, context):
    json_bucket = event["Records"][0]["s3"]["bucket"]["name"]
    json_key = urllib.parse.unquote(event["Records"][0]["s3"]["object"]["key"])
    rdf_bucket = get_required_lambda_env_var("RDF_BUCKET")
    rdf_key = ".".join(json_key.split(".")[:-1]) + ".rdf.gz"
    session = boto3.Session()
    s3_client = session.client("s3")

    logger = Logger()
    with logger.bind(json_bucket=json_bucket, json_key=json_key):
        graph_pkg = graph_pkg_from_s3(
            s3_client=s3_client, json_bucket=json_bucket, json_key=json_key
        )

    with logger.bind(rdf_bucket=rdf_bucket, rdf_key=rdf_key):
        logger.info(event=LogEvent.WriteToS3Start)
        with io.BytesIO() as rdf_bytes_buf:
            with gzip.GzipFile(fileobj=rdf_bytes_buf, mode="wb") as gz:
                graph_pkg.graph.serialize(gz)
            rdf_bytes_buf.flush()
            rdf_bytes_buf.seek(0)
            s3_client.upload_fileobj(rdf_bytes_buf, rdf_bucket, rdf_key)
            s3_client.put_object_tagging(
                Bucket=rdf_bucket,
                Key=rdf_key,
                Tagging={
                    "TagSet": [
                        {"Key": "name", "Value": graph_pkg.name},
                        {"Key": "version", "Value": graph_pkg.version},
                        {"Key": "start_time", "Value": str(graph_pkg.start_time)},
                        {"Key": "end_time", "Value": str(graph_pkg.end_time)},
                    ]
                },
            )
        logger.info(event=LogEvent.WriteToS3End)


def graph_set_from_filepath(filepath: str) -> GraphSet:
    with open(filepath, "r") as fp:
        graph_set_dict = json.load(fp)
    return GraphSet.from_dict(data=graph_set_dict)


def graph_from_filepath(filepath: str) -> Graph:
    graph_set = graph_set_from_filepath(filepath)
    return graph_set.to_rdf()


def main(argv=None):
    import argparse

    if argv is None:
        argv = sys.argv[1:]
    parser = argparse.ArgumentParser()
    parser.add_argument("infile", type=str)
    parser.add_argument("outfile", type=str)
    args_ns = parser.parse_args(argv)
    graph = graph_from_filepath(args_ns.infile)
    graph.serialize(args_ns.outfile)


if __name__ == "__main__":
    sys.exit(main())



"""Resource for VPCs"""
from typing import Type

from botocore.client import BaseClient

from altimeter.aws.resource.resource_spec import ListFromAWSResult
from altimeter.aws.resource.ec2 import EC2ResourceSpec
from altimeter.core.graph.field.scalar_field import ScalarField
from altimeter.core.graph.field.tags_field import TagsField
from altimeter.core.graph.schema import Schema


class VPCResourceSpec(EC2ResourceSpec):
    """Resource for VPCs"""

    type_name = "vpc"
    schema = Schema(
        ScalarField("IsDefault"), ScalarField("CidrBlock"), ScalarField("State"), TagsField()
    )

    @classmethod
    def list_from_aws(
        cls: Type["VPCResourceSpec"], client: BaseClient, account_id: str, region: str
    ) -> ListFromAWSResult:
        """Return a dict of dicts of the format:

            {'vpc_1_arn': {vpc_1_dict},
             'vpc_2_arn': {vpc_2_dict},
             ...}

        Where the dicts represent results from describe_vpcs."""
        vpcs = {}
        paginator = client.get_paginator("describe_vpcs")
        for resp in paginator.paginate():
            for vpc in resp.get("Vpcs", []):
                resource_arn = cls.generate_arn(
                    account_id=account_id, region=region, resource_id=vpc["VpcId"]
                )
                vpcs[resource_arn] = vpc
        return ListFromAWSResult(resources=vpcs)



"""Resource representing an unscanned AWS Account"""
from typing import List, Type

from botocore.client import BaseClient

from altimeter.core.graph.link.links import SimpleLink
from altimeter.core.graph.link.base import Link
from altimeter.core.graph.schema import Schema
from altimeter.aws.resource.resource_spec import ScanGranularity, ListFromAWSResult, AWSResourceSpec
from altimeter.core.resource.resource import Resource
from altimeter.aws.scan.aws_accessor import AWSAccessor


class UnscannedAccountResourceSpec(AWSResourceSpec):
    """Resource representing an unscanned AWS Account"""

    type_name = "unscanned-account"
    service_name = "null"
    scan_granularity = ScanGranularity.ACCOUNT
    schema = Schema()

    @classmethod
    def create_resource(
        cls: Type["UnscannedAccountResourceSpec"], account_id: str, errors: List[str]
    ) -> Resource:
        links: List[Link] = []
        links.append(SimpleLink(pred="account_id", obj=account_id))
        for error in errors:
            link = SimpleLink(pred="error", obj=error)
            links.append(link)
        return Resource(
            resource_id=cls.generate_arn(resource_id=account_id),
            type_name=cls.get_full_type_name(),
            links=links,
        )

    @classmethod
    def get_full_type_name(cls: Type["UnscannedAccountResourceSpec"]) -> str:
        return f"{cls.provider_name}:{cls.type_name}"

    @classmethod
    def list_from_aws(
        cls: Type["UnscannedAccountResourceSpec"], client: BaseClient, account_id: str, region: str
    ) -> ListFromAWSResult:
        """List resources from AWS using client."""

    @classmethod
    def generate_arn(
        cls: Type["UnscannedAccountResourceSpec"],
        resource_id: str,
        account_id: str = "",
        region: str = "",
    ) -> str:
        """Generate an ARN for this resource"""
        return f"arn:aws::::account/{resource_id}"

    @classmethod
    def scan(
        cls: Type["UnscannedAccountResourceSpec"], scan_accessor: AWSAccessor
    ) -> List[Resource]:
        raise NotImplementedError(f"{cls.__name__} is not a scannable ResourceSpec class.")



"""Resource for IAM Users"""
import copy
from typing import Any, List, Dict, Type

from botocore.client import BaseClient
from botocore.exceptions import ClientError

from altimeter.aws.resource.resource_spec import ListFromAWSResult
from altimeter.aws.resource.iam import IAMResourceSpec
from altimeter.core.graph.field.dict_field import AnonymousDictField, DictField, EmbeddedDictField
from altimeter.core.graph.field.list_field import ListField
from altimeter.core.graph.field.scalar_field import ScalarField
from altimeter.core.graph.schema import Schema


class IAMUserResourceSpec(IAMResourceSpec):
    """Resource for IAM Users"""

    type_name = "user"
    schema = Schema(
        ScalarField("UserName", "name"),
        ScalarField("UserId"),
        ScalarField("CreateDate"),
        ScalarField("PasswordLastUsed", optional=True),
        ListField(
            "AccessKeys",
            EmbeddedDictField(
                ScalarField("AccessKeyId"),
                ScalarField("Status"),
                ScalarField("CreateDate"),
                AnonymousDictField("AccessKeyLastUsed", ScalarField("LastUsedDate", optional=True)),
            ),
            optional=True,
            alti_key="access_key",
        ),
        DictField(
            "LoginProfile",
            ScalarField("CreateDate"),
            ScalarField("PasswordResetRequired"),
            optional=True,
        ),
        ListField(
            "MfaDevices",
            EmbeddedDictField(ScalarField("SerialNumber"), ScalarField("EnableDate")),
            optional=True,
        ),
    )

    @classmethod
    def list_from_aws(
        cls: Type["IAMUserResourceSpec"], client: BaseClient, account_id: str, region: str
    ) -> ListFromAWSResult:
        """Return a dict of dicts of the format:

            {'user_1_arn': {user_1_dict},
             'user_2_arn': {user_2_dict},
             ...}

        Where the dicts represent results from list_users and additional info per user from
        list_targets_by_user."""
        users = {}
        paginator = client.get_paginator("list_users")
        for resp in paginator.paginate():
            for user in resp.get("Users", []):
                resource_arn = user["Arn"]
                user_name = user["UserName"]
                access_keys_paginator = client.get_paginator("list_access_keys")
                access_keys: List[Dict[str, Any]] = []
                for access_keys_resp in access_keys_paginator.paginate(UserName=user_name):
                    for resp_access_key in access_keys_resp["AccessKeyMetadata"]:
                        access_key = copy.deepcopy(resp_access_key)
                        access_key_id = access_key["AccessKeyId"]
                        last_used_resp = client.get_access_key_last_used(AccessKeyId=access_key_id)
                        access_key["AccessKeyLastUsed"] = last_used_resp["AccessKeyLastUsed"]
                        access_keys.append(access_key)
                user["AccessKeys"] = access_keys
                mfa_devices_paginator = client.get_paginator("list_mfa_devices")
                mfa_devices: List[Dict[str, Any]] = []
                for mfa_devices_resp in mfa_devices_paginator.paginate(UserName=user_name):
                    mfa_devices += mfa_devices_resp["MFADevices"]
                    user["MfaDevices"] = mfa_devices
                try:
                    login_profile_resp = client.get_login_profile(UserName=user_name)
                    user["LoginProfile"] = login_profile_resp["LoginProfile"]
                except ClientError as c_e:
                    if "NoSuchEntity" not in str(c_e):
                        raise c_e
                users[resource_arn] = user
        return ListFromAWSResult(resources=users)



"""Exceptions related ton Graphing"""
from altimeter.core.exceptions import AltimeterException


class LinkParseException(AltimeterException):
    """Error parsing Links from JSON."""


class SchemaParseException(AltimeterException):
    """Schema.parse error"""


class UnmergableDuplicateResourceIdsFoundException(AltimeterException):
    """Duplicate resource ids were found in a GraphSet and are not mergable"""


class ListFieldSourceKeyNotFoundException(AltimeterException):
    """The source_key of a ListField was not found"""


class ListFieldValueNotAListException(AltimeterException):
    """A ListField does not contain a list"""


class GraphSetOrphanedReferencesException(AltimeterException):
    """A GraphSet contained orphaned references, for instance a ResourceLink referring to
    an id not present in the GraphSet."""


class UnmergableGraphSetsException(AltimeterException):
    """GraphSets are unable to be merged."""



"""A ScanManifest defines the output of a complete scan."""

from dataclasses import dataclass
from typing import Any, Dict, List


@dataclass(frozen=True)
class ScanManifest:
    """A ScanManifest defines the output of a complete scan. It contains pointers to the
    per-account scan result artifacts and summaries of what was scanned, errors which occurred,
    scan datetime and api call statistics.

    Args:
        scanned_accounts: Dict of account_ids to account detail dicts for scanned accounts
        master_artifact: artifact containing complete graph json
        artifacts: list of artifacts, one per account
        errors: Dict of account_ids to list of errors encountered during scan
        unscanned_accounts: Dict of account_ids to account detail dicts for unscanned accounts
        api_call_stats: api call stats for this scan
        start_time: epoch timestamp of scan start time
        end_time: epoch timestamp of scan end time
    """

    scanned_accounts: List[Dict[str, str]]
    master_artifact: str
    artifacts: List[str]
    errors: Dict[str, List[str]]
    unscanned_accounts: List[Dict[str, str]]
    api_call_stats: Dict[str, Any]
    start_time: int
    end_time: int

    def to_dict(self) -> Dict[str, Any]:
        """Generate a dict representation of this ScanManifest.

        Returns:
            dict representation of this ScanManifest
        """
        return {
            "scanned_accounts": self.scanned_accounts,
            "master_artifact": self.master_artifact,
            "artifacts": self.artifacts,
            "errors": self.errors,
            "unscanned_accounts": self.unscanned_accounts,
            "start_time": self.start_time,
            "end_time": self.end_time,
            "api_call_stats": self.api_call_stats,
        }



"""Resource representing an AWS Organizational Unit."""
from typing import Any, Dict, List, Type

from botocore.client import BaseClient

from altimeter.aws.resource.resource_spec import ListFromAWSResult
from altimeter.aws.resource.organizations import OrganizationsResourceSpec
from altimeter.aws.resource.organizations.org import OrgResourceSpec
from altimeter.core.graph.field.scalar_field import ScalarField
from altimeter.core.graph.field.resource_link_field import ResourceLinkField
from altimeter.core.graph.schema import Schema


class OUResourceSpec(OrganizationsResourceSpec):
    """Resource representing an AWS OU."""

    type_name = "ou"
    schema = Schema(
        ScalarField("Path"), ResourceLinkField("OrganizationArn", OrgResourceSpec, value_is_id=True)
    )

    @classmethod
    def get_full_type_name(cls: Type["OUResourceSpec"]) -> str:
        return f"{cls.provider_name}:{cls.type_name}"

    @classmethod
    def list_from_aws(
        cls: Type["OUResourceSpec"], client: BaseClient, account_id: str, region: str
    ) -> ListFromAWSResult:
        """Return a dict of dicts of the format:

            {'ou_1_arn': {ou_1_dict},
             'ou_2_arn': {ou_2_dict},
             ...}

        Where the dicts represent results from list_organizational_units_for_parent
        with some additional info 'Path') tagged on."""
        org_resp = client.describe_organization()
        org_arn = org_resp["Organization"]["Arn"]
        ous = {}
        paginator = client.get_paginator("list_roots")
        for resp in paginator.paginate():
            for root in resp["Roots"]:
                root_id, root_arn = root["Id"], root["Arn"]
                root_path = f"/{root['Name']}"
                ous[root_arn] = root
                ous[root_arn]["OrganizationArn"] = org_arn
                ous[root_arn]["Path"] = root_path
                ou_details = cls._recursively_get_ou_details_for_parent(
                    client=client, parent_id=root_id, parent_path=root_path
                )
                for ou_detail in ou_details:
                    arn = ou_detail["Arn"]
                    ou_detail["OrganizationArn"] = org_arn
                    ous[arn] = ou_detail
        return ListFromAWSResult(resources=ous)

    @classmethod
    def _recursively_get_ou_details_for_parent(
        cls: Type["OUResourceSpec"], client: BaseClient, parent_id: str, parent_path: str
    ) -> List[Dict[str, Any]]:
        ous = []
        paginator = client.get_paginator("list_organizational_units_for_parent")
        for resp in paginator.paginate(ParentId=parent_id):
            for ou in resp["OrganizationalUnits"]:
                ou_id = ou["Id"]
                path = f"{parent_path}/{ou['Name']}"
                ou["Path"] = path
                ous.append(ou)
                ous += cls._recursively_get_ou_details_for_parent(
                    client=client, parent_id=ou_id, parent_path=path
                )
        return ous



"""Resource for Clusters"""
from typing import Type

from botocore.client import BaseClient
from botocore.exceptions import ClientError

from altimeter.aws.resource.resource_spec import ListFromAWSResult
from altimeter.aws.resource.eks import EKSResourceSpec
from altimeter.core.graph.field.scalar_field import ScalarField
from altimeter.core.graph.schema import Schema


class EKSClusterResourceSpec(EKSResourceSpec):
    """Resource for Clusters"""

    type_name = "cluster"
    schema = Schema(ScalarField("Name"),)

    @classmethod
    def list_from_aws(
        cls: Type["EKSClusterResourceSpec"], client: BaseClient, account_id: str, region: str
    ) -> ListFromAWSResult:
        """Return a dict of dicts of the format:

            {'cluster_1_arn': {cluster_1_dict},
             'cluster_2_arn': {cluster_2_dict},
             ...}

        Where the dicts represent results from list_clusters."""
        clusters = {}
        try:
            paginator = client.get_paginator("list_clusters")
            for resp in paginator.paginate():
                for cluster_name in resp.get("clusters", []):
                    resource_arn = cls.generate_arn(
                        account_id=account_id, region=region, resource_id=cluster_name
                    )
                    clusters[resource_arn] = {"Name": cluster_name}
        except ClientError as c_e:
            response_error = getattr(c_e, "response", {}).get("Error", {})
            error_code = response_error.get("Code", "")
            if error_code != "AccessDeniedException":
                raise c_e
            error_msg = response_error.get("Message", "")
            if error_msg != f"Account {account_id} is not authorized to use this service":
                raise c_e
        return ListFromAWSResult(resources=clusters)



"""Function for encoding JSON with datetimes."""
from datetime import datetime
from typing import Any


def json_encoder(obj: Any) -> Any:
    """json encoder function supporting datetime serialization.

    Args:
        obj: object to encode to JSON

    Returns:
        json encoded data
    """
    if isinstance(obj, datetime):
        return obj.isoformat()
    raise TypeError("Type {} not serializable".format(type(obj)))



"""Resource for RDS Snapshot"""
from typing import Type

from botocore.client import BaseClient

from altimeter.aws.resource.resource_spec import ListFromAWSResult
from altimeter.aws.resource.ec2.vpc import VPCResourceSpec
from altimeter.aws.resource.rds import RDSResourceSpec
from altimeter.aws.resource.rds.instance import RDSInstanceResourceSpec
from altimeter.aws.resource.kms.key import KMSKeyResourceSpec
from altimeter.core.graph.field.resource_link_field import TransientResourceLinkField
from altimeter.core.graph.field.scalar_field import ScalarField
from altimeter.core.graph.schema import Schema


class RDSSnapshotResourceSpec(RDSResourceSpec):
    """Resource for RDS Snapshot"""

    type_name = "snapshot"
    schema = Schema(
        ScalarField("DBSnapshotIdentifier"),
        ScalarField("SnapshotCreateTime", optional=True),
        ScalarField("Engine"),
        ScalarField("AllocatedStorage"),
        ScalarField("Status"),
        TransientResourceLinkField("VpcId", VPCResourceSpec, optional=True),
        ScalarField("InstanceCreateTime"),
        ScalarField("SnapshotType"),
        ScalarField("PercentProgress"),
        ScalarField("Region", optional=True),
        ScalarField("Encrypted"),
        TransientResourceLinkField("KmsKeyID", KMSKeyResourceSpec, optional=True),
        ScalarField("Timezone", optional=True),
        ScalarField("IAMDatabaseAuthenticationEnabled"),
        TransientResourceLinkField("DBInstanceArn", RDSInstanceResourceSpec, value_is_id=True),
    )

    @classmethod
    def generate_instance_arn(
        cls: Type["RDSSnapshotResourceSpec"], account_id: str, region: str, resource_id: str
    ) -> str:
        """Generate an ARN for this resource, e.g. arn:aws:rds:<region>:<account>:db:<name>
        """
        return ":".join(
            (
                "arn",
                cls.provider_name,
                cls.service_name,
                region,
                account_id,
                cls.type_name,
                resource_id,
            )
        )

    @classmethod
    def list_from_aws(
        cls: Type["RDSSnapshotResourceSpec"], client: BaseClient, account_id: str, region: str
    ) -> ListFromAWSResult:
        dbinstances = {}
        paginator = client.get_paginator("describe_db_snapshots")
        for resp in paginator.paginate():
            for db in resp.get("DBSnapshots", []):
                resource_arn = db["DBSnapshotArn"]
                db["DBInstanceArn"] = cls.generate_instance_arn(
                    account_id, region, db["DBInstanceIdentifier"]
                )
                dbinstances[resource_arn] = db
        return ListFromAWSResult(resources=dbinstances)



"""Resource for KMSKeys"""
from typing import Type

from botocore.client import BaseClient

from altimeter.aws.resource.resource_spec import ListFromAWSResult
from altimeter.aws.resource.kms import KMSResourceSpec
from altimeter.core.graph.field.scalar_field import ScalarField
from altimeter.core.graph.schema import Schema


class KMSKeyResourceSpec(KMSResourceSpec):
    """Resource for KMS Keys"""

    type_name = "key"
    schema = Schema(ScalarField("KeyId"))

    @classmethod
    def list_from_aws(
        cls: Type["KMSKeyResourceSpec"], client: BaseClient, account_id: str, region: str
    ) -> ListFromAWSResult:
        """Return a dict of dicts of the format:

            {'key_1_arn': {key_1_dict},
             'key_2_arn': {key_2_dict},
             ...}

        Where the dicts represent results from list_keys."""
        keys = {}
        paginator = client.get_paginator("list_keys")
        for resp in paginator.paginate():
            for key in resp.get("Keys", []):
                resource_arn = key["KeyArn"]
                keys[resource_arn] = key
        return ListFromAWSResult(resources=keys)



"""Resource for Transit Gateways"""
from typing import Any, Dict, List, Type

from botocore.client import BaseClient

from altimeter.aws.resource.resource_spec import ListFromAWSResult
from altimeter.aws.resource.ec2 import EC2ResourceSpec
from altimeter.aws.resource.ec2.vpc import VPCResourceSpec
from altimeter.core.graph.field.dict_field import EmbeddedDictField
from altimeter.core.graph.field.list_field import ListField
from altimeter.core.graph.field.resource_link_field import ResourceLinkField
from altimeter.core.graph.field.scalar_field import ScalarField
from altimeter.core.graph.field.tags_field import TagsField
from altimeter.core.graph.schema import Schema


class TransitGatewayResourceSpec(EC2ResourceSpec):
    """Resource for TransitGateways"""

    type_name = "transit-gateway"
    schema = Schema(
        ScalarField("OwnerId"),
        ScalarField("State"),
        ListField(
            "VPCAttachments",
            EmbeddedDictField(
                ScalarField("CreationTime"),
                ScalarField("State"),
                ResourceLinkField("ResourceId", VPCResourceSpec),
            ),
            optional=True,
            alti_key="vpc_attachment",
        ),
        TagsField(),
    )

    @classmethod
    def list_from_aws(
        cls: Type["TransitGatewayResourceSpec"], client: BaseClient, account_id: str, region: str
    ) -> ListFromAWSResult:
        """Return a dict of dicts of the format:

            {'tgw_1_arn': {tgw_1_dict},
             'tgw_2_arn': {tgw_2_dict},
             ...}

        Where the dicts represent results from describe_transit_gateways."""
        tgws = {}
        paginator = client.get_paginator("describe_transit_gateways")
        tgw_filters = [{"Name": "owner-id", "Values": [account_id]}]
        for resp in paginator.paginate(Filters=tgw_filters):
            for tgw in resp["TransitGateways"]:
                resource_arn = tgw["TransitGatewayArn"]
                vpc_attachments: List[Dict[str, Any]] = []
                vpc_attachments_paginator = client.get_paginator(
                    "describe_transit_gateway_attachments"
                )
                vpc_filters = [
                    {"Name": "transit-gateway-id", "Values": [tgw["TransitGatewayId"]]},
                    {"Name": "resource-type", "Values": ["vpc"]},
                ]
                for vpc_attachments_resp in vpc_attachments_paginator.paginate(Filters=vpc_filters):
                    vpc_attachments += vpc_attachments_resp["TransitGatewayAttachments"]
                tgw["VPCAttachments"] = vpc_attachments
                tgws[resource_arn] = tgw
        return ListFromAWSResult(resources=tgws)



"""Base class for IAM resources."""

from altimeter.aws.resource.resource_spec import ScanGranularity, AWSResourceSpec


class IAMResourceSpec(AWSResourceSpec):
    """Base class for IAM resources."""

    service_name = "iam"
    scan_granularity = ScanGranularity.ACCOUNT



"""A ResourceSpec defines how an individual resource (e.g. EC2 Instances) JSON
is converted into graph data.  It contains a Schema which contains a set of
Fields defining the transformation."""
import abc
from collections import defaultdict
import inspect
from typing import Any, DefaultDict, Dict, List, Set, Type

from altimeter.core.resource.exceptions import (
    MultipleResourceSpecClassesFoundException,
    ResourceSpecClassNotFoundException,
)
from altimeter.core.graph.exceptions import UnmergableDuplicateResourceIdsFoundException
from altimeter.core.graph.link.base import Link
from altimeter.core.graph.schema import Schema
from altimeter.core.resource.resource import Resource


class ResourceSpec(abc.ABC):
    """A ResourceSpec defines how an individual resource (e.g. EC2 Instances) JSON
    is converted into graph data.  It contains a Schema which contains a set of
    Fields defining the transformation."""

    schema: Schema = Schema()
    type_name: str = ""
    # A list of ResourceSpec classes which are allowed to clobber/overwrite this one if
    # duplicate resource_ids are found.  The primary use case for this is to allow
    # UnscannedAccountResourceSpec generated resources to overwrite AccountResourceSpecs.
    allow_clobber: List[Type] = []

    def __init_subclass__(cls: Type["ResourceSpec"], **kwargs: Any) -> None:
        if not inspect.isabstract(cls):
            for required in ("schema", "type_name"):
                if not getattr(cls, required):
                    raise TypeError(
                        f"Can not instantiate {cls.__name__} without {required} attribute."
                    )
        return super().__init_subclass__()

    @classmethod
    @abc.abstractmethod
    def scan(cls: Type["ResourceSpec"], scan_accessor: Any) -> List[Resource]:
        """Scan for this ResourceSpec using scan_accessor and return a list of Resource objects

        Args:
            scan_accessor: scan accessor object for accessing required APIs

        Returns:
            List of Resource objects
        """

    @classmethod
    def get_full_type_name(cls: Type["ResourceSpec"]) -> str:
        """Get the fully qualified type name for this class.  Generally this is something like
        aws:ec2:type_name.

        Returns:
            full resource type name string
        """
        return cls.type_name

    @classmethod
    def generate_id(
        cls: Type["ResourceSpec"], short_resource_id: str, context: Dict[str, Any]
    ) -> str:
        """Generate a full id for this type given a short_resource_id.

        Args:
            short_resource_id: short resource id for this resource
            context: contains auxiliary information which can be passed through the parse process.

        Returns:
            full resource id string
        """
        return f"{cls.type_name}:{short_resource_id}"

    @staticmethod
    def get_by_class_name(class_name: str) -> Type["ResourceSpec"]:
        """Get a ResourceSpec class by class name.

        Args:
            class_name: class name to match

        Returns:
            ResourceSpec subclass

        Raises:
            MultipleResourceSpecClassesFoundException if more than one match was found.
                this indicates a code bug.
            ResourceSpecClassNotFoundException if no ResourceSpec class was found.
        """
        subclasses = get_concrete_subclasses(ResourceSpec)
        candidates = []
        for subclass in subclasses:
            if subclass.__name__ == class_name:
                candidates.append(subclass)
        if len(candidates) == 1:
            return candidates.pop()
        if len(candidates) > 1:
            raise MultipleResourceSpecClassesFoundException(
                f"Multiple ResourceSpec subclasses found named {class_name}: {candidates}"
            )
        raise ResourceSpecClassNotFoundException(
            f"No ResourceSpec subclass found named {class_name}"
        )

    @staticmethod
    def get_by_full_type_name(type_name: str) -> List[Type["ResourceSpec"]]:
        """Get a ResourceSpec classes by full_type name.

        Args:
           type_name: type name of ResourceSpec to find

        Returns:
            list of ResourceSpec classes matching type_name

        Raises:
            ResourceSpecClassNotFoundException if no ResourceSpec classes could be
            found matching type_name.
        """
        subclasses: List[Type[ResourceSpec]] = get_concrete_subclasses(ResourceSpec)
        candidates = []
        for subclass in subclasses:
            if subclass.get_full_type_name() == type_name:
                candidates.append(subclass)
        if candidates:
            return candidates
        raise ResourceSpecClassNotFoundException(
            f"No ResourceSpec subclass found with full_type_name {type_name}"
        )

    @staticmethod
    def merge_resources(resource_id: str, resources: List[Resource]) -> Resource:
        """Merge multiple resources with the same id into one.  This is permissible
        in two cases:

            1) If all resources have the same value for 'get_full_type' and no key/values
            conflict they will be merged into a single resource by combining all key/values
            into a single resource

            2) If all resources do not have the same value for 'get_full_type', if classes
            have the allow_clobber attribute set depending on the values a resource may be
            created.

        Args:
             resource_id: common resource id
             resources: list of Resources to merge

        Returns:
            merged Resource object

        Raises:
            UnmergableDuplicateResourceIdsFoundException if resources could not be merged.
        """
        full_type_names = {resource.type_name for resource in resources}
        if len(full_type_names) > 1:
            # in this case conflicting resources have different full_type_names, this
            # can be a permissible if allow_clobber is used. Here we determine this by
            # building a dict where the keys are resource spec classes which have
            # appeared in the duplicate resources allow_clobber list and the
            # values are the classes where they appeared.  If at the end in
            # this dict there are any values which match the complete list of
            # spec classes, we use the resource of that type
            spec_classes_resources: DefaultDict[Type[ResourceSpec], List[Resource]] = defaultdict(
                list
            )
            allow_clobber_classes_classes: DefaultDict[
                Type[ResourceSpec], Set[Type[ResourceSpec]]
            ] = defaultdict(set)
            for resource in resources:
                full_type_name = resource.type_name
                spec_classes: List[Type[ResourceSpec]] = ResourceSpec.get_by_full_type_name(
                    full_type_name
                )
                for spec_class in spec_classes:
                    spec_classes_resources[spec_class].append(resource)
                    for allow_clobber_class in spec_class.allow_clobber:
                        allow_clobber_classes_classes[allow_clobber_class].add(spec_class)
            all_spec_classes = set(spec_classes_resources.keys())
            winning_class = None
            for allow_clobber_class, classes in allow_clobber_classes_classes.items():
                if len(classes) == len(all_spec_classes) - 1:
                    winning_class = (all_spec_classes - classes).pop()
            if not winning_class:
                raise UnmergableDuplicateResourceIdsFoundException(
                    (
                        f"Multiple resources for {resource_id} with "
                        f"different types that aren't clobberable: "
                        f"{[resource.to_dict() for resource in resources]}"
                    )
                )
            resources = spec_classes_resources[winning_class]
        full_type_names = {resource.type_name for resource in resources}
        merged_resource_type_name = full_type_names.pop()
        merged_link_keys_links: Dict[str, Link] = {}
        for duplicate_resource in resources:
            for link in duplicate_resource.links:
                duplicate_link = merged_link_keys_links.get(link.pred)
                if duplicate_link:
                    if duplicate_link.field_type != link.field_type:
                        raise UnmergableDuplicateResourceIdsFoundException(
                            f"Conflicting link types {link.field_type}, {duplicate_link.field_type} found in duplicate #s {resources}"
                        )
                    if duplicate_link.obj != link.obj:
                        raise UnmergableDuplicateResourceIdsFoundException(
                            f"Conflicting link values {link.obj}, {duplicate_link.obj} found in duplicate #s {resources}"
                        )
                else:
                    merged_link_keys_links[link.pred] = link
        merged_links = list(merged_link_keys_links.values())
        return Resource(
            resource_id=resource_id, type_name=merged_resource_type_name, links=merged_links
        )


def get_concrete_subclasses(cls: Type[Any]) -> List[Type]:
    """Get all concrete subclasses of a class.

    Args:
        cls: class to find subclasses of

    Returns:
        list of classes which are concrete subclasses of cls
    """
    all_subclasses = cls.__subclasses__()
    concrete_subclasses = [sc for sc in all_subclasses if not inspect.isabstract(sc)]
    for subclass in all_subclasses:
        concrete_subclasses += get_concrete_subclasses(subclass)
    return concrete_subclasses



"""Classes for ArtifactReaders. An ArtifactReader reads a scan artifact dict
from something - e.g. a file, s3 key, etc."""
import abc
import io
import json
from typing import Any, Dict, Tuple

import boto3

from altimeter.core.log import LogEvent, Logger
from altimeter.core.artifact_io.exceptions import InvalidS3URIException


class ArtifactReader(abc.ABC):
    """ArtifactReaders read JSON artifacts from locations - e.g. s3, filesystem, etc."""

    @abc.abstractmethod
    def read_artifact(self, artifact_path: str) -> Dict[str, Any]:
        """Read an artifact

        Args:
            artifact_path: path to artifact to read

        Returns:
            artifact content
        """


class FileArtifactReader(ArtifactReader):
    """ArtifactReader to read from the filesystem"""

    def read_artifact(self, artifact_path: str) -> Dict[str, Any]:
        """Read an artifact

        Args:
            artifact_path: filesystem path to artifact

        Returns:
            artifact content
        """
        logger = Logger()
        with logger.bind(artifact_path=artifact_path):
            logger.info(event=LogEvent.ReadFromFSStart)
            with open(artifact_path, "r") as artifact_fp:
                data = json.load(artifact_fp)
            logger.info(event=LogEvent.ReadFromFSEnd)
            return data


class S3ArtifactReader(ArtifactReader):
    """ArtifactReader to read from S3"""

    def read_artifact(self, artifact_path: str) -> Dict[str, Any]:
        """Read an artifact

        Args:
            artifact_path: s3 uri to artifact. s3://bucket/key/path

        Returns:
            artifact content
        """
        bucket, key = parse_s3_uri(artifact_path)
        session = boto3.Session()
        s3_client = session.client("s3")
        logger = Logger()
        with io.BytesIO() as artifact_bytes_buf:
            with logger.bind(bucket=bucket, key=key):
                logger.info(event=LogEvent.ReadFromS3Start)
                s3_client.download_fileobj(bucket, key, artifact_bytes_buf)
                artifact_bytes_buf.flush()
                artifact_bytes_buf.seek(0)
                artifact_bytes = artifact_bytes_buf.read()
                logger.info(event=LogEvent.ReadFromS3End)
                artifact_str = artifact_bytes.decode("utf-8")
                artifact_dict = json.loads(artifact_str)
                return artifact_dict


def parse_s3_uri(uri: str) -> Tuple[str, str]:
    """Parse an s3 uri (s3://bucket/key/path) into bucket and key parts

    Args:
        uri: s3 uri (s3://bucket/key/path)

    Returns:
        Tuple of (bucket, key)

    Raises:
        :class:`InvalidS3URIException` if the uri
        argument is not a valid S3 URI.
    """
    s3_uri_prefix = "s3://"
    if not uri.startswith(s3_uri_prefix):
        raise InvalidS3URIException("S3 URIs should begin with 's3://")
    uri = uri.rstrip("/")
    parts = uri[len(s3_uri_prefix) :].split("/")
    if len(parts) < 2:
        raise InvalidS3URIException(f"{uri} missing key portion")
    bucket = parts[0]
    if not bucket:
        raise InvalidS3URIException(f"Bad bucket portion in uri {uri}")
    key_parts = [part.rstrip("/ ") for part in parts[1:]]
    if not all(key_parts):
        raise InvalidS3URIException(f"Bad key portion in uri {uri}")
    key = "/".join(key_parts)
    return bucket, key



import abc
import inspect
from typing import Any, Dict

from rdflib import BNode, Namespace, Graph

from altimeter.core.graph.node_cache import NodeCache


class Link(abc.ABC):
    """A Link represents the predicate-object portion of a triple.
    Links in general have a field_type which describes the nature of the relationship.

    Args:
        pred: predicate portion of the triple this Link represents
        obj: object portion of the triple this Link represents.
    """

    field_type: str = ""

    def __init__(self, pred: str, obj: Any):
        self.pred = pred
        self.obj = obj

    def __init_subclass__(cls, **kwargs: Any) -> None:
        if not inspect.isabstract(cls):
            for required in ("field_type",):
                if not getattr(cls, required):
                    raise TypeError(
                        f"Can not instantiate {cls.__name__} without {required} attribute."
                    )
        return super().__init_subclass__()

    def to_dict(self) -> Dict[str, str]:
        """Return a dictionary representation of this Link

        Returns:
            Dict representation of this Link
        """
        return {"pred": self.pred, "obj": self.obj, "type": self.field_type}

    @classmethod
    def from_dict(cls, pred: str, obj: Any) -> Any:
        """Create a Link object from a dict

        Args:
            pred: Link predicate
            obj: Link object

        Returns:
            Link subclass object
        """
        return cls(pred=pred, obj=obj)

    @abc.abstractmethod
    def to_rdf(
        self, subj: BNode, namespace: Namespace, graph: Graph, node_cache: NodeCache
    ) -> None:
        """Graph this link on a BNode in a Graph using a given Namespace to create the full
        predicate.

        Args:
             subj: subject portion of triple - graph this link's pred, obj against it.
             namespace: RDF namespace to use for this triple's predicate
             graph: RDF graph
             node_cache: NodeCache to use to find cached nodes.
        """



"""AWSResourceSpec is a subclass of ResourceSpec which is used to define
ResourceSpecs for AWS resources"""
import abc
import inspect
from dataclasses import dataclass
from enum import Enum, auto
from typing import Dict, Any, List, Tuple, Type

from botocore.client import BaseClient
from botocore.exceptions import ClientError

from altimeter.aws.scan.aws_accessor import AWSAccessor
from altimeter.core.graph.exceptions import SchemaParseException
from altimeter.core.graph.link.links import ResourceLinkLink
from altimeter.core.resource.resource import Resource
from altimeter.core.resource.resource_spec import ResourceSpec

AWS_API_IGNORE_ERRORS = frozenset(
    ("NotSignedUp", "OptInRequired", "SubscriptionRequiredException", "InvalidAction")
)


class ScanGranularity(Enum):
    """ScanGranularities are attached to AWSResourceSpecs and define how resources are scanned."""

    REGION = auto()  # scanned in every region
    ACCOUNT = auto()  # scanned in only one region


@dataclass(frozen=True)
class ListFromAWSResult:
    """Result of a list_from_aws call. Contains a list of resources represented as a dict
    of arns to resource details

    Args:
        resources: Dict of resource ids to resource dicts
    """

    resources: Dict[str, Dict[str, Any]]


class AWSResourceSpec(ResourceSpec):
    """AWSResourceSpec is a subclass of ResourceSpec which is used to define
    ResourceSpecs for AWS resources"""

    provider_name: str = "aws"
    service_name: str = ""
    scan_granularity: ScanGranularity = ScanGranularity.REGION
    region_whitelist: Tuple[str, ...] = ()

    def __init_subclass__(cls: Type["AWSResourceSpec"], **kwargs: Any) -> None:
        if not inspect.isabstract(cls):
            for required in ("service_name",):
                if not getattr(cls, required):
                    raise TypeError(
                        f"Can not instantiate {cls.__name__} without {required} attribute."
                    )
        return super().__init_subclass__(**kwargs)

    @classmethod
    def get_full_type_name(cls: Type["AWSResourceSpec"]) -> str:
        """Get the fully qualified type name for this class, generally something like
        aws:ec2:instance, aws:iam:role, etc.

        Returns:
            string of full type name, generally something like "aws:ec2:instance"
        """
        return f"{cls.provider_name}:{cls.service_name}:{cls.type_name}"

    @classmethod
    def get_client_name(cls: Type["AWSResourceSpec"]) -> str:
        """Get the boto3 client name to be used for scanning resources of this type.
        Generally this is the same as cls.service_name but in some cases it is not.

        Returns:
             string of boto3 client name for cls.service
        """
        return cls.service_name

    @classmethod
    def generate_id(
        cls: Type["AWSResourceSpec"], short_resource_id: str, context: Dict[str, Any]
    ) -> str:
        """Generate a full id (arn) given a short resource id.

        Args:
            short_resource_id: last portion of an aws arn - e.g. i-1234, ami-abcd, etc.
            context: dict containing account_id, region

        Returns:
            string containing resource arn.
        """

        return cls.generate_arn(
            account_id=context.get("account_id", ""),
            region=context.get("region", ""),
            resource_id=short_resource_id,
        )

    @classmethod
    def generate_arn(
        cls: Type["AWSResourceSpec"], resource_id: str, account_id: str = "", region: str = "",
    ) -> str:
        """Generate an ARN for this resource

        Args:
            account_id: resource account id
            region: resource region
            resource_id: resource id

        Returns:
            string containing resource arn.
        """
        return (
            ":".join(
                ("arn", cls.provider_name, cls.service_name, region, account_id, cls.type_name)
            )
            + f"/{resource_id}"
        )

    @classmethod
    def skip_resource_scan(
        cls: Type["AWSResourceSpec"], client: BaseClient, account_id: str, region: str
    ) -> bool:
        """Return a bool indicating whether this resource class scan should be skipped.
        Args:
            client: boto3 client
            account_id: account id
            region: aws region

        Returns:
            True if this resource should be skipped.
        """
        return False

    @classmethod
    @abc.abstractmethod
    def list_from_aws(
        cls: Type["AWSResourceSpec"], client: BaseClient, account_id: str, region: str
    ) -> ListFromAWSResult:
        """Return a ListFromAWSResult object by calling the appropriate
        AWS API calls to list/describe the resource represented by this class.

        Args:
            client: boto3 Client
            account_id: aws account id
            region: aws region

        Returns:
            ListFromAWSResult object
        """

    @classmethod
    def scan(cls: Type["AWSResourceSpec"], scan_accessor: AWSAccessor) -> List[Resource]:
        """Scan this ResourceSpec

       Args:
           scan_accessor: AWSAccessor object to use for api access

        Returns:
            List of Resource objects
        """
        context = {"account_id": scan_accessor.account_id, "region": scan_accessor.region}
        list_from_aws_result = cls._list_from_aws(scan_accessor)
        resources = cls._list_from_aws_result_to_resources(
            list_from_aws_result=list_from_aws_result, context=context
        )
        return resources

    @classmethod
    def _list_from_aws(
        cls: Type["AWSResourceSpec"], scan_accessor: AWSAccessor
    ) -> ListFromAWSResult:
        try:
            resource_client = scan_accessor.client(cls.get_client_name())
            if cls.skip_resource_scan(
                client=resource_client,
                account_id=scan_accessor.account_id,
                region=scan_accessor.region,
            ):
                return ListFromAWSResult(resources={})
            return cls.list_from_aws(
                resource_client, scan_accessor.account_id, scan_accessor.region
            )
        except ClientError as c_e:
            response_error = getattr(c_e, "response", {}).get("Error", {})
            error_code = response_error.get("Code", "")
            if error_code not in AWS_API_IGNORE_ERRORS:
                raise c_e
            return ListFromAWSResult(resources={})

    @classmethod
    def _list_from_aws_result_to_resources(
        cls: Type["AWSResourceSpec"],
        list_from_aws_result: ListFromAWSResult,
        context: Dict[str, str],
    ) -> List[Resource]:
        resources: List[Resource] = []
        for arn, resource_dict in list_from_aws_result.resources.items():
            try:
                links = cls.schema.parse(resource_dict, context)
            except Exception as ex:
                raise SchemaParseException(
                    (
                        f"Error parsing {cls.__name__} : "
                        f"{arn}:\n\nData: {resource_dict}\n\nError: {ex}"
                    )
                )
            resource_account_id = arn.split(":")[4]
            if resource_account_id:
                if resource_account_id != "aws":
                    account_link = ResourceLinkLink(
                        pred="account", obj=f"arn:aws::::account/{resource_account_id}"
                    )
                    links.append(account_link)
                    resource_region_name = arn.split(":")[3]
                    if resource_region_name:
                        region_link = ResourceLinkLink(
                            pred="region",
                            obj=f"arn:aws:::{resource_account_id}:region/{resource_region_name}",
                        )
                        links.append(region_link)
            resource = Resource(resource_id=arn, type_name=cls.get_full_type_name(), links=links)
            resources.append(resource)
        return resources



"""Resource for SecurityGroups"""
import ipaddress
from typing import Type

from botocore.client import BaseClient

from altimeter.aws.resource.resource_spec import ListFromAWSResult
from altimeter.aws.resource.ec2 import EC2ResourceSpec
from altimeter.core.graph.field.dict_field import EmbeddedDictField
from altimeter.core.graph.field.list_field import ListField
from altimeter.core.graph.field.resource_link_field import ResourceLinkField
from altimeter.core.graph.field.scalar_field import ScalarField
from altimeter.core.graph.field.tags_field import TagsField
from altimeter.core.graph.schema import Schema


class SecurityGroupResourceSpec(EC2ResourceSpec):
    """Resource for SecurityGroups"""

    type_name = "security-group"
    schema = Schema(
        ScalarField("GroupName", "name"),
        ListField(
            "IpPermissions",
            EmbeddedDictField(
                ScalarField("IpProtocol"),
                ScalarField("FromPort", default_value=0),
                ScalarField("ToPort", default_value=65535),
                ListField(
                    "IpRanges",
                    EmbeddedDictField(
                        ScalarField("CidrIp"), ScalarField("FirstIp"), ScalarField("LastIp")
                    ),
                    alti_key="ip_range",
                    optional=True,
                ),
                ListField(
                    "Ipv6Ranges",
                    EmbeddedDictField(
                        ScalarField("CidrIpv6"), ScalarField("FirstIp"), ScalarField("LastIp")
                    ),
                    alti_key="ipv6_range",
                    optional=True,
                ),
                ListField(
                    "PrefixListIds", EmbeddedDictField(ScalarField("PrefixListId")), optional=True
                ),
                ListField(
                    "UserIdGroupPairs",
                    EmbeddedDictField(
                        ResourceLinkField("GroupId", "SecurityGroupResourceSpec"),
                        ScalarField("UserId", alti_key="account_id"),
                        ScalarField("PeeringStatus", optional=True),
                        ScalarField("VpcId", optional=True),
                        ScalarField("VpcPeeringConnectionId", optional=True),
                    ),
                    alti_key="user_id_group_pairs",
                ),
            ),
            alti_key="ingress_rule",
        ),
        ListField(
            "IpPermissionsEgress",
            EmbeddedDictField(
                ScalarField("IpProtocol"),
                ScalarField("FromPort", default_value=0),
                ScalarField("ToPort", default_value=65535),
                ListField(
                    "IpRanges",
                    EmbeddedDictField(
                        ScalarField("CidrIp"), ScalarField("FirstIp"), ScalarField("LastIp")
                    ),
                    alti_key="ip_range",
                    optional=True,
                ),
                ListField(
                    "Ipv6Ranges",
                    EmbeddedDictField(
                        ScalarField("CidrIpv6"), ScalarField("FirstIp"), ScalarField("LastIp")
                    ),
                    alti_key="ipv6_range",
                    optional=True,
                ),
                ListField(
                    "PrefixListIds", EmbeddedDictField(ScalarField("PrefixListId")), optional=True
                ),
                ListField(
                    "UserIdGroupPairs",
                    EmbeddedDictField(
                        ResourceLinkField("GroupId", "SecurityGroupResourceSpec"),
                        ScalarField("UserId", alti_key="account_id"),
                        ScalarField("PeeringStatus", optional=True),
                        ScalarField("VpcId", optional=True),
                        ScalarField("VpcPeeringConnectionId", optional=True),
                    ),
                    alti_key="user_id_group_pairs",
                ),
            ),
            alti_key="egress_rule",
        ),
        TagsField(),
    )

    @classmethod
    def list_from_aws(
        cls: Type["SecurityGroupResourceSpec"], client: BaseClient, account_id: str, region: str
    ) -> ListFromAWSResult:
        """Return a dict of dicts of the format:

            {'security_group_1_arn': {security_group_1_dict},
             'security_group_2_arn': {security_group_2_dict},
             ...}

        Where the dicts represent results from describe_subnets."""
        security_groups = {}
        paginator = client.get_paginator("describe_security_groups")
        for resp in paginator.paginate():
            for security_group in resp.get("SecurityGroups", []):
                resource_arn = cls.generate_arn(
                    account_id=account_id, region=region, resource_id=security_group["GroupId"]
                )
                for ingress_rule in security_group.get("IpPermissions", []):
                    for ip_range in ingress_rule.get("IpRanges", []):
                        cidr = ip_range["CidrIp"]
                        ipv4_network = ipaddress.IPv4Network(cidr, strict=False)
                        first_ip, last_ip = int(ipv4_network[0]), int(ipv4_network[-1])
                        ip_range["FirstIp"] = first_ip
                        ip_range["LastIp"] = last_ip
                    for ip_range in ingress_rule.get("Ipv6Ranges", []):
                        cidr = ip_range["CidrIpv6"]
                        ipv6_network = ipaddress.IPv6Network(cidr, strict=False)
                        first_ip, last_ip = int(ipv6_network[0]), int(ipv6_network[-1])
                        ip_range["FirstIp"] = first_ip
                        ip_range["LastIp"] = last_ip
                for egress_rule in security_group.get("IpPermissionsEgress", []):
                    for ip_range in egress_rule.get("IpRanges", []):
                        cidr = ip_range["CidrIp"]
                        ipv4_network = ipaddress.IPv4Network(cidr, strict=False)
                        first_ip, last_ip = int(ipv4_network[0]), int(ipv4_network[-1])
                        ip_range["FirstIp"] = first_ip
                        ip_range["LastIp"] = last_ip
                    for ip_range in egress_rule.get("Ipv6Ranges", []):
                        cidr = ip_range["CidrIpv6"]
                        ipv6_network = ipaddress.IPv6Network(cidr, strict=False)
                        first_ip, last_ip = int(ipv6_network[0]), int(ipv6_network[-1])
                        ip_range["FirstIp"] = first_ip
                        ip_range["LastIp"] = last_ip
                security_groups[resource_arn] = security_group
        return ListFromAWSResult(resources=security_groups)



"""Contains various utility functions useful in Lambdas."""
import os
from typing import Any, Dict

from altimeter.core.awslambda.exceptions import (
    RequiredEnvironmentVariableNotPresentException,
    RequiredEventVariableNotPresentException,
)


def get_required_lambda_env_var(key: str) -> str:
    """Get a variable from os.environ.

    Args:
        key: Key to look up in the os environment.

    Returns:
        String value for the given key

    Raises:
        RequiredEnvironmentVariableNotPresentException if key is not present.
    """
    value = os.environ.get(key)
    if value is None:
        raise RequiredEnvironmentVariableNotPresentException(f"Missing required env var {key}")
    return value


def get_required_lambda_event_var(event: Dict[str, Any], key: str) -> Any:
    """Get a variable from a lambda event dict.

    Args:
        event: Lambda event dict
        key: Key to look up in the event

    Returns:
        String value for the given key

    Raises:
        RequiredEventVariableNotPresentException if key is not in event
    """
    value = event.get(key)
    if value is None:
        raise RequiredEventVariableNotPresentException(
            f"Missing required event var {key} in {event}"
        )
    return value



"""AWSResourceSpec subclass for eks resources."""

from altimeter.aws.resource.resource_spec import AWSResourceSpec


class EKSResourceSpec(AWSResourceSpec):
    """AWSResourceSpec subclass for eks resources."""

    service_name = "eks"



"""Resource for VPC Endpoints"""
from typing import Type

from botocore.client import BaseClient

from altimeter.aws.resource.resource_spec import ListFromAWSResult
from altimeter.aws.resource.ec2 import EC2ResourceSpec
from altimeter.aws.resource.ec2.vpc import VPCResourceSpec
from altimeter.core.graph.field.scalar_field import ScalarField
from altimeter.core.graph.field.resource_link_field import ResourceLinkField
from altimeter.core.graph.field.tags_field import TagsField
from altimeter.core.graph.schema import Schema


class VpcEndpointResourceSpec(EC2ResourceSpec):
    """Resource for VPC Endpoints"""

    type_name = "vpc-endpoint"
    schema = Schema(
        ScalarField("VpcEndpointType"),
        ScalarField("ServiceName"),
        ScalarField("State"),
        ResourceLinkField("VpcId", VPCResourceSpec),
        TagsField(),
    )

    @classmethod
    def list_from_aws(
        cls: Type["VpcEndpointResourceSpec"], client: BaseClient, account_id: str, region: str
    ) -> ListFromAWSResult:
        """Return a dict of dicts of the format:

            {'vpc_endpoint_1_arn': {vpc_endpoint_1_dict},
             'vpc_endpoint_1_arn': {vpc_endpoint_2_dict},
             ...}

        Where the dicts represent results from describe_vpc_endpoints."""
        endpoints = {}
        paginator = client.get_paginator("describe_vpc_endpoints")
        for resp in paginator.paginate():
            for endpoint in resp.get("VpcEndpoints", []):
                resource_arn = cls.generate_arn(
                    account_id=account_id, region=region, resource_id=endpoint["VpcEndpointId"]
                )
                endpoints[resource_arn] = endpoint
        return ListFromAWSResult(resources=endpoints)



"""Resource for EC2 Network Interfaces."""
from typing import Type

from botocore.client import BaseClient

from altimeter.aws.resource.resource_spec import ListFromAWSResult
from altimeter.aws.resource.ec2 import EC2ResourceSpec
from altimeter.aws.resource.ec2.vpc import VPCResourceSpec
from altimeter.core.graph.field.dict_field import AnonymousDictField
from altimeter.core.graph.field.resource_link_field import ResourceLinkField
from altimeter.core.graph.field.scalar_field import ScalarField
from altimeter.core.graph.schema import Schema


class EC2NetworkInterfaceResourceSpec(EC2ResourceSpec):
    """Resource for EC2NetworkInterfaces"""

    type_name = "network-interface"
    schema = Schema(
        AnonymousDictField(
            "Association",
            ScalarField("PublicDnsName", optional=True),
            ScalarField("PublicIp", optional=True),
            optional=True,
        ),
        ScalarField("Description"),
        ScalarField("InterfaceType"),
        ScalarField("MacAddress"),
        ScalarField("PrivateDnsName", optional=True),
        ScalarField("PrivateIpAddress", optional=True),
        ScalarField("Status"),
        ResourceLinkField("VpcId", VPCResourceSpec, optional=True),
    )

    @classmethod
    def list_from_aws(
        cls: Type["EC2NetworkInterfaceResourceSpec"],
        client: BaseClient,
        account_id: str,
        region: str,
    ) -> ListFromAWSResult:
        """Return a dict of dicts of the format:

            {'network_interface_1_arn': {network_interface_1_dict},
             'network_interface_2_arn': {network_interface_2_dict},
             ...}

        Where the dicts represent results from describe_network_interfaces."""
        paginator = client.get_paginator("describe_network_interfaces")
        interfaces = {}
        for resp in paginator.paginate():
            for interface in resp.get("NetworkInterfaces", []):
                resource_arn = cls.generate_arn(
                    account_id=account_id,
                    region=region,
                    resource_id=interface["NetworkInterfaceId"],
                )
                interfaces[resource_arn] = interface
        return ListFromAWSResult(resources=interfaces)



#!/usr/bin/env python3
import json

from altimeter.core.artifact_io.writer import S3ArtifactWriter
from altimeter.aws.scan.account_scanner import AccountScanner
from altimeter.aws.scan.account_scan_plan import AccountScanPlan
from altimeter.core.json_encoder import json_encoder
from altimeter.core.awslambda import get_required_lambda_event_var


def lambda_handler(event, context):
    account_scan_plan_dict = get_required_lambda_event_var(event, "account_scan_plan")
    account_scan_plan = AccountScanPlan.from_dict(account_scan_plan_dict)
    json_bucket = get_required_lambda_event_var(event, "json_bucket")
    key_prefix = get_required_lambda_event_var(event, "key_prefix")
    scan_sub_accounts = get_required_lambda_event_var(event, "scan_sub_accounts")

    artifact_writer = S3ArtifactWriter(bucket=json_bucket, key_prefix=key_prefix)
    account_scanner = AccountScanner(
        account_scan_plan=account_scan_plan,
        artifact_writer=artifact_writer,
        scan_sub_accounts=scan_sub_accounts,
    )
    scan_results_dict = account_scanner.scan()
    scan_results_str = json.dumps(scan_results_dict, default=json_encoder)
    json_results = json.loads(scan_results_str)
    return json_results



"""A MultiHopAccessor contains a list of AccessSteps defining how to gain access to an account via
role assumption(s)."""
from dataclasses import asdict, dataclass, field
from typing import Any, Dict, List, Optional, Type

import boto3

from altimeter.aws.auth.cache import AWSCredentials, AWSCredentialsCache
from altimeter.core.log import LogEvent, Logger


@dataclass(frozen=True)
class AccessStep:
    """Represents a single access step to get to an account.

    Args:
        role_name: role name for this step
        account_id: account_id for this step. If empty this step is assumed to be the last
                    in a chain of multiple AccessSteps
        external_id: external_id to use for access (if needed).
    """

    role_name: str
    account_id: Optional[str] = field(default=None)
    external_id: Optional[str] = field(default=None)

    def __str__(self) -> str:
        account = self.account_id if self.account_id else "target"
        return f"{self.role_name}@{account}"

    @classmethod
    def from_dict(cls: Type["AccessStep"], data: Dict[str, Any]) -> "AccessStep":
        return cls(**data)


@dataclass(frozen=True)
class MultiHopAccessor:
    """A MultiHopAccessor contains a list of AccessSteps defining how to gain access to an account
    via role assumption(s).

    Args:
        role_session_name: role session name to use for session creation.
        access_steps: list of AccessSteps defining how to access a final
                      destination account.
    """

    role_session_name: str
    access_steps: List[AccessStep]

    def __post_init__(self) -> None:
        if not self.access_steps:
            raise ValueError("One or more access steps must be specified")
        for access_step in self.access_steps[:-1]:
            if not access_step.account_id:
                raise ValueError(
                    "Non-final AccessStep of a MultiHopAccessor must specify an account_id"
                )
        if self.access_steps[-1].account_id:
            raise ValueError(
                "The last AccessStep of a MultiHopAccessor must not specify account_id"
            )

    def get_session(
        self,
        account_id: str,
        region_name: Optional[str] = None,
        credentials_cache: Optional[AWSCredentialsCache] = None,
    ) -> boto3.Session:
        """Get a session for an account_id by iterating through the :class:`.AccessStep`s
        of this :class:`.MultiHopAccessor`.

        Args:
             account_id: account to access
             region_name: region to use during session creation.

        Returns:
            boto3 Session for accessing account_id
        """
        logger = Logger()
        cws = boto3.Session(region_name=region_name)
        for access_step in self.access_steps:
            access_account_id = access_step.account_id if access_step.account_id else account_id
            role_name = access_step.role_name
            external_id = access_step.external_id
            session = None
            if credentials_cache is not None:
                session = credentials_cache.get(
                    account_id=access_account_id,
                    role_name=role_name,
                    role_session_name=self.role_session_name,
                    region_name=region_name,
                )
            if session is None:
                logger.debug(event=LogEvent.AuthToAccountStart)
                sts_client = cws.client("sts")
                role_arn = f"arn:aws:iam::{access_account_id}:role/{role_name}"
                assume_args = {"RoleArn": role_arn, "RoleSessionName": self.role_session_name}
                if external_id:
                    assume_args["ExternalId"] = external_id

                assume_resp = sts_client.assume_role(**assume_args)
                creds = assume_resp["Credentials"]
                expiration_datetime = creds["Expiration"]
                credentials = AWSCredentials(
                    access_key_id=creds["AccessKeyId"],
                    secret_access_key=creds["SecretAccessKey"],
                    session_token=creds["SessionToken"],
                    expiration=int(expiration_datetime.timestamp()),
                )
                session = boto3.Session(
                    aws_access_key_id=creds["AccessKeyId"],
                    aws_secret_access_key=creds["SecretAccessKey"],
                    aws_session_token=creds["SessionToken"],
                    region_name=region_name,
                )
                if credentials_cache is not None:
                    credentials_cache.put(
                        credentials=credentials,
                        account_id=access_account_id,
                        role_name=role_name,
                        role_session_name=self.role_session_name,
                    )
                logger.debug(event=LogEvent.AuthToAccountEnd)
            cws = session
        return session

    def __str__(self) -> str:
        return f'accessor:{self.role_session_name}:{",".join([str(access_step) for access_step in self.access_steps])}'

    def to_dict(self) -> Dict[str, Any]:
        return asdict(self)

    @classmethod
    def from_dict(cls: Type["MultiHopAccessor"], data: Dict[str, Any]) -> "MultiHopAccessor":
        role_session_name = data.get("role_session_name")
        if not role_session_name:
            raise ValueError(f"Expected key 'role_session_name' in {data} with non-empty value")
        access_step_dicts = data.get("access_steps", [])
        access_steps = [
            AccessStep.from_dict(access_step_dict) for access_step_dict in access_step_dicts
        ]
        return cls(role_session_name=role_session_name, access_steps=access_steps,)



"""Resource for IAM Policies"""
from typing import Type

from botocore.client import BaseClient

from altimeter.aws.resource.resource_spec import ListFromAWSResult
from altimeter.aws.resource.iam import IAMResourceSpec
from altimeter.aws.resource.util import policy_doc_dict_to_sorted_str
from altimeter.core.graph.field.scalar_field import ScalarField
from altimeter.core.graph.schema import Schema


class IAMPolicyResourceSpec(IAMResourceSpec):
    """Resource for user-managed IAM Policies"""

    type_name = "policy"
    schema = Schema(
        ScalarField("PolicyName", "name"),
        ScalarField("PolicyId"),
        ScalarField("DefaultVersionId"),
        ScalarField("DefaultVersionPolicyDocumentText"),
    )

    @classmethod
    def list_from_aws(
        cls: Type["IAMPolicyResourceSpec"], client: BaseClient, account_id: str, region: str
    ) -> ListFromAWSResult:
        """Return a dict of dicts of the format:

            {'role_1_arn': {role_1_dict},
             'role_2_arn': {role_2_dict},
             ...}

        Where the dicts represent results from list_policies and additional info per role from
        list_targets_by_role."""
        policies = {}
        paginator = client.get_paginator("list_policies")

        for resp in paginator.paginate(Scope="Local"):
            for policy in resp.get("Policies", []):
                resource_arn = policy["Arn"]
                default_policy_version = policy["DefaultVersionId"]
                policy_version_resp = client.get_policy_version(
                    PolicyArn=resource_arn, VersionId=default_policy_version
                )
                default_policy_version_document_text = policy_version_resp["PolicyVersion"][
                    "Document"
                ]
                policy["DefaultVersionPolicyDocumentText"] = policy_doc_dict_to_sorted_str(
                    default_policy_version_document_text
                )
                policies[resource_arn] = policy
        return ListFromAWSResult(resources=policies)


class IAMAWSManagedPolicyResourceSpec(IAMResourceSpec):
    """Resource for AWS-managed IAM Policies"""

    type_name = "policy"
    schema = Schema(ScalarField("PolicyName", "name"), ScalarField("PolicyId"))

    @classmethod
    def list_from_aws(
        cls: Type["IAMAWSManagedPolicyResourceSpec"],
        client: BaseClient,
        account_id: str,
        region: str,
    ) -> ListFromAWSResult:
        """Return a dict of dicts of the format:

            {'role_1_arn': {role_1_dict},
             'role_2_arn': {role_2_dict},
             ...}

        Where the dicts represent results from list_policies and additional info per role from
        list_targets_by_role."""
        policies = {}
        paginator = client.get_paginator("list_policies")

        for resp in paginator.paginate(Scope="AWS", OnlyAttached=True):
            for policy in resp.get("Policies", []):
                resource_arn = policy["Arn"]
                policies[resource_arn] = policy
        return ListFromAWSResult(resources=policies)



"""A Schema consists of a list of Fields which define how to parse an arbitrary dictionary
into a list of Links."""
from typing import Any, Dict, List

from altimeter.core.graph.field.base import Field
from altimeter.core.graph.link.base import Link


class Schema:
    """A Schema consists of a list of Fields which define how to parse an arbitrary dictionary
    into a list of :class:`altimeter.core.graph.links.Link`.
    The schema method performs translation to :class:`altimeter.core.graph.links.Link`.

    Args:
        fields: fields for this Schema.
    """

    def __init__(self, *fields: Field) -> None:
        self.fields = fields

    def parse(self, data: Dict[str, Any], context: Dict[str, Any]) -> List[Link]:
        """Parse this schema into a list of Links

        Args:
            data: raw data to parse
            context: contains auxiliary information which can be passed through the parse process.

        Returns:
            A list of :class:`altimeter.core.graph.links.Link` .
        """
        links: List[Any] = []
        for field in self.fields:
            links += field.parse(data, context)
        return links



"""ResourceSpec classes for elbv2 resources."""
from altimeter.aws.resource.resource_spec import AWSResourceSpec


class ELBV2ResourceSpec(AWSResourceSpec):
    """Abstract base for ResourceSpec classes for elbv2 resources."""

    service_name = "elbv2"

    @classmethod
    def get_client_name(cls) -> str:
        return "elbv2"



"""Resource for Account Password Policy"""
from typing import Type

from botocore.client import BaseClient

from altimeter.aws.resource.resource_spec import ListFromAWSResult
from altimeter.aws.resource.iam import IAMResourceSpec
from altimeter.core.graph.field.scalar_field import ScalarField
from altimeter.core.graph.schema import Schema


class IAMAccountPasswordPolicyResourceSpec(IAMResourceSpec):
    """Resource for Account Password Policy"""

    DEFAULT_PASSWORD_POLICY_NAME = "default"

    type_name = "account-password-policy"
    schema = Schema(
        ScalarField("MinimumPasswordLength"),
        ScalarField("RequireSymbols"),
        ScalarField("RequireNumbers"),
        ScalarField("RequireUppercaseCharacters"),
        ScalarField("RequireLowercaseCharacters"),
        ScalarField("AllowUsersToChangePassword"),
        ScalarField("ExpirePasswords"),
        ScalarField("MaxPasswordAge", optional=True),
        ScalarField("PasswordReusePrevention", optional=True),
        ScalarField("HardExpiry", optional=True),
    )

    @classmethod
    def list_from_aws(
        cls: Type["IAMAccountPasswordPolicyResourceSpec"],
        client: BaseClient,
        account_id: str,
        region: str,
    ) -> ListFromAWSResult:
        """Return a dict of dicts of the format:

            {'account_password_policy_1_arn': {account_password_policy_1_dict}}

        Where the dicts represent results from get_account_password_policy."""
        password_policies = {}
        try:
            resp = client.get_account_password_policy()
        except client.exceptions.NoSuchEntityException:
            resp = {}  # Indicates no policy is set for the account

        policy = resp.get("PasswordPolicy", {})

        if policy:
            resource_arn = cls.generate_arn(
                account_id=account_id, resource_id=cls.DEFAULT_PASSWORD_POLICY_NAME
            )
            password_policies[resource_arn] = policy
        return ListFromAWSResult(resources=password_policies)



"""Resource for Instance Profiles"""
from typing import Type

from botocore.client import BaseClient

from altimeter.aws.resource.resource_spec import ListFromAWSResult
from altimeter.aws.resource.iam import IAMResourceSpec
from altimeter.aws.resource.iam.role import IAMRoleResourceSpec
from altimeter.core.graph.field.dict_field import AnonymousEmbeddedDictField
from altimeter.core.graph.field.list_field import AnonymousListField
from altimeter.core.graph.field.resource_link_field import ResourceLinkField
from altimeter.core.graph.field.scalar_field import ScalarField
from altimeter.core.graph.schema import Schema


class InstanceProfileResourceSpec(IAMResourceSpec):
    """Resource for Instance Profiles"""

    type_name = "instance-profile"
    schema = Schema(
        ScalarField("InstanceProfileName", alti_key="name"),
        AnonymousListField(
            "Roles",
            AnonymousEmbeddedDictField(
                ResourceLinkField("Arn", IAMRoleResourceSpec, value_is_id=True)
            ),
        ),
    )

    @classmethod
    def list_from_aws(
        cls: Type["InstanceProfileResourceSpec"], client: BaseClient, account_id: str, region: str
    ) -> ListFromAWSResult:
        """Return a dict of dicts of the format:

            {'instance_profile_1_arn': {instance_profile_1_dict},
             'instance_profile_2_arn': {instance_profile_2_dict},
             ...}

        Where the dicts represent results from list_instance_profiles."""
        paginator = client.get_paginator("list_instance_profiles")
        instance_profiles = {}
        for resp in paginator.paginate():
            for instance_profile in resp.get("InstanceProfiles", []):
                resource_arn = instance_profile["Arn"]
                instance_profiles[resource_arn] = instance_profile
        return ListFromAWSResult(resources=instance_profiles)



"""An AccountScanPlan defines how to scan a set of accounts."""
from dataclasses import dataclass
from typing import Any, Dict, List, Type

from altimeter.aws.auth.accessor import Accessor


@dataclass(frozen=True)
class AccountScanPlan:
    """An AccountScanPlan defines how to scan a set of accounts.

    Arguments:
        account_ids: account ids to scan
        regions: regions to scan
        accessor: Accessor to use to access the accounts
    """

    account_ids: List[str]
    regions: List[str]
    accessor: Accessor

    def to_dict(self) -> Dict[str, Any]:
        """Generate a dict representation of this AccountScanPlan.

        Returns:
            dict representation of this AccountScanPlan
        """
        return {
            "account_ids": self.account_ids,
            "regions": list(self.regions),
            "accessor": self.accessor.to_dict(),
        }

    @classmethod
    def from_dict(
        cls: Type["AccountScanPlan"], account_scan_plan_dict: Dict[str, Any]
    ) -> "AccountScanPlan":
        """Create an AccountScanPlan from a dict

        Args:
           account_scan_plan_dict: dict of AccountScanPlan data

        Returns:
            AccountScanPlan object
        """
        account_ids = account_scan_plan_dict["account_ids"]
        regions = account_scan_plan_dict["regions"]
        accessor_dict = account_scan_plan_dict["accessor"]
        accessor = Accessor.from_dict(accessor_dict)
        return cls(account_ids=account_ids, regions=regions, accessor=accessor)

    def to_batches(self, max_accounts: int) -> List["AccountScanPlan"]:
        """Break this AccountScanPlan into multiple AccountScanPlans with a max of
        max_accounts account_ids per plan"""
        account_id_batches = [
            self.account_ids[n : n + max_accounts]
            for n in range(0, len(self.account_ids), max_accounts)
        ]
        return [
            AccountScanPlan(
                account_ids=account_id_batch, regions=self.regions, accessor=self.accessor
            )
            for account_id_batch in account_id_batches
        ]



from typing import Optional, List, Dict, Any, Type

from rdflib import Namespace, Graph, BNode, RDF, Literal

from altimeter.core.graph.link.base import Link
from altimeter.core.graph.link.links import link_from_dict
from altimeter.core.graph.node_cache import NodeCache


class Resource:
    """A Resource defines a single scanned resource which is directly translatable to a graph
    node.  It contains an id, type name and list of Links.

    Args:
         resource_id: id of this resource
         type_name: type name of this resource
         links: List of Links for this resource
    """

    def __init__(self, resource_id: str, type_name: str, links: Optional[List[Link]] = None):
        self.resource_id = resource_id
        self.type_name = type_name
        self.links = [] if links is None else links

    def to_dict(self) -> Dict[str, Any]:
        """Generate a dict representation of this Resource.

        Returns:
            dict representation of this Resource
        """
        scan_json: Dict[str, Any] = {"type": self.type_name}
        if self.links:
            scan_json["links"] = [link.to_dict() for link in self.links]
        return scan_json

    @classmethod
    def from_dict(
        cls: Type["Resource"], resource_id: str, resource_data: Dict[str, Any]
    ) -> "Resource":
        """Create an instances of this class from a resource_id and resource_data dict
        as generated by to_dict.

        Args:
            resource_id: resource id
            resource_data: dict of data for this resource

        Returns:
            Resource object
        """
        type_name = resource_data["type"]
        links: List[Link] = []
        for link in resource_data.get("links", []):
            links.append(link_from_dict(link))
        return cls(resource_id=resource_id, type_name=type_name, links=links)

    def to_rdf(self, namespace: Namespace, graph: Graph, node_cache: NodeCache) -> None:
        """Graph this Resource as a BNode on a Graph.

        Args:
            namespace: RDF namespace to use for predicates and objects when graphing
                       this resource's links
            graph: RDF graph
            node_cache: NodeCache to use for any cached BNode lookups
        """
        node = node_cache.setdefault(self.resource_id, BNode())
        graph.add((node, RDF.type, getattr(namespace, self.type_name)))
        graph.add((node, getattr(namespace, "id"), Literal(self.resource_id)))
        for link in self.links:
            link.to_rdf(subj=node, namespace=namespace, graph=graph, node_cache=node_cache)



"""Base class for S3 resources."""
from altimeter.aws.resource.resource_spec import AWSResourceSpec


class S3ResourceSpec(AWSResourceSpec):
    """Base class for S3 resources."""

    service_name = "s3"



"""AWS scan settings

Attributes:
    GRAPH_NAME: name for the AWS graph.
    GRAPH_VERSION: AWS graph version.  Generally this should change very rarely.
"""
import os

GRAPH_NAME: str = "alti"
GRAPH_VERSION: str = "1"

# The following variables control scan concurrency.
#
# The execution model of aws2json is as such:
#
# aws2json.py (single process)
#       AccountScan (up to MAX_<exectype>_ACCOUNT_SCAN_THREADS concurrently)
#             AWSScanMuxer (up to MAX_SCAN_THREADS concurrently)
#
# In general, the maximum number of concurrent scan operations is:
#
#   MAX_<exectype>_ACCOUNT_SCAN_THREADS * MAX_SCAN_THREADS
#
# Note that each of these variables is overridable via env vars, see os.environ.get calls below.

# in lambda, number of AccountScan lambdas to run concurrently
MAX_LAMBDA_ACCOUNT_SCAN_THREADS = int(os.environ.get("MAX_MUXER_THREADS", 64))
# in lambda, the number of accounts to scan in each AccountScan lambda
MAX_LAMBDA_ACCOUNTS_PER_THREAD = int(os.environ.get("MAX_ACCOUNTS_PER_THREAD", 1))

# in local usage, number of AccountScans to run concurrently
MAX_LOCAL_ACCOUNT_SCAN_THREADS = int(os.environ.get("MAX_MUXER_THREADS", 4))
# in lcoal usage, the number of accounts to scan in each AccountScan
MAX_LOCAL_ACCOUNTS_PER_THREAD = int(os.environ.get("MAX_ACCOUNTS_PER_THREAD", 1))

# the number of scan threads to spawn in each account scan
MAX_ACCOUNT_SCANNER_THREADS = int(os.environ.get("MAX_ACCOUNT_SCANNER_THREADS", 64))

# for ACCOUNT granularity resources, this is the list of preferred scan regions
PREFERRED_ACCOUNT_SCAN_REGIONS = ("us-west-1", "us-west-2", "us-east-1", "us-east-2")



"""Base class for GuardDuty resources."""

from altimeter.aws.resource.resource_spec import AWSResourceSpec


class GuardDutyResourceSpec(AWSResourceSpec):
    """Base class for GuardDuty resources."""

    service_name = "guardduty"



"""Base class for KMS resources."""
from altimeter.aws.resource.resource_spec import AWSResourceSpec


class KMSResourceSpec(AWSResourceSpec):
    """Base class for KMS resources."""

    service_name = "kms"



"""Multilevel Counter is a counter which allows counting categorized items in a hierarchical
fashion."""

from collections import defaultdict
from typing import Any, DefaultDict, Dict, Type, Union


class MultilevelCounter:
    """Multilevel Counter is a counter which allows counting categorized items in a hierarchical
    fashion.

    Examples:
        For example if you are counting API calls to made to AWS accounts/regions/services,
        the following could be used to count:

            >>> mc = MultilevelCounter()
            >>> # call to account 123, region us-east-1, service ec2
            >>> mc.increment('123', 'us-east-1', 'ec2')
            >>> # another call to account 123, region us-east-1, service ec2
            >>> mc.increment('123', 'us-east-1', 'ec2')
            >>> # call to account 123, region us-east-1 service s3
            >>> mc.increment('123', 'us-east-1', 's3')
            >>> # call to account 456, region eu-west-1, service ecs
            >>> mc.increment('456', 'eu-west-1', 'ecs')

            >>> # The output below reflects that we made a total of 4 calls,
            >>> # 3 to account 123, 1 to account 456, and so on.
            >>> import json
            >>> print(json.dumps(mc.to_dict(), indent=2))
            {
              "count": 4,
              "123": {
                "count": 3,
                "us-east-1": {
                  "count": 3,
                  "ec2": {
                    "count": 2
                  },
                  "s3": {
                    "count": 1
                  }
                }
              },
              "456": {
                "count": 1,
                "eu-west-1": {
                  "count": 1,
                  "ecs": {
                    "count": 1
                  }
                }
              }
            }
    """

    def __init__(self) -> None:
        self.count = 0
        self.multilevel_counters: DefaultDict[str, MultilevelCounter] = defaultdict(
            MultilevelCounter
        )

    def increment(self, *categories: str) -> None:
        """Increment a given category tuple.

        Args:
             categories: ordered categories to increment.
        """
        self.count += 1
        if categories:
            category = categories[0]
            self.multilevel_counters[category].increment(*categories[1:])

    def merge(self, other: "MultilevelCounter") -> None:
        """Merge another MultilevelCounter into this counter.

        Args:
            other: MultilevelCounter to merge into self.
        """
        self.count += other.count
        for other_cat, other_stat in other.multilevel_counters.items():
            self.multilevel_counters[other_cat].merge(other_stat)

    def to_dict(self) -> Dict[str, Union[int, Dict[str, Any]]]:
        """Convert the contents of this MultilevelCounter to a dict.

        Returns:
            dict representation of this counter.
        """
        data: Dict[str, Union[int, Dict[str, Any]]] = {"count": self.count}
        for cat, stat in self.multilevel_counters.items():
            data[cat] = stat.to_dict()
        return data

    @classmethod
    def from_dict(
        cls: Type["MultilevelCounter"], stats_data: Dict[str, Any]
    ) -> "MultilevelCounter":
        """Create a MultilevelCounter from a dict.

        Args:
            stats_data: dict representing MultilevelCounter data

        Returns:
            MultilevelCounter
        """
        count = 0
        multilevel_counters: DefaultDict[str, MultilevelCounter]
        multilevel_counters = defaultdict(MultilevelCounter)
        for key, val in stats_data.items():
            if key == "count":
                count = val
            else:
                multilevel_counters[key] = cls.from_dict(val)
        obj = cls()
        obj.count = count
        obj.multilevel_counters = multilevel_counters
        return obj
